/**
 * Copyright © 2012-2020 by Data Virtuality GmbH
 */
package org.teiid.metadata;

import com.datavirtuality.dv.core.interfaces.DVCaller;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.NavigableMap;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.ConcurrentSkipListMap;
import java.util.concurrent.TimeUnit;
import org.teiid.adminapi.DVMetadataCache;
import org.teiid.metadata.odata.ODataMetadata;
import org.teiid.metadata.odata.ODataSchemaBuilder;

/**
 * @author Dmitry Pogorelov
 * @date 2020.10.13
 */
public class DVMetadataRecordHolder implements DVMetadataCache {
	private NavigableMap<String, AbstractMetadataRecord> uuidHolder = new ConcurrentSkipListMap<>(String.CASE_INSENSITIVE_ORDER);
	private Timer timer = new Timer("UUID Holder Cleaner", true);
	private static final int THRESHOLD_TO_REMOVE = 10;
	public static final long REFRESH_TIMEOUT = TimeUnit.SECONDS.toMillis(60);
	private volatile boolean testMode;
	private Object lock = new Object();
	private ODataMetadata oDataMetadata;

	// for debug purposes
	private NavigableMap<String, AbstractMetadataRecord> uuidHolder2 = new ConcurrentSkipListMap<>(String.CASE_INSENSITIVE_ORDER);
	private static boolean test;

	public DVMetadataRecordHolder(ODataMetadata oDataMetadata, boolean runCleaner) {
		this.oDataMetadata = oDataMetadata;
		if (runCleaner)
			timer.schedule(new UUIDHolderCleaner(this), 0, REFRESH_TIMEOUT);
	}

	public DVMetadataRecordHolder(String vdbName, boolean runCleaner) {
		this(new ODataMetadata(vdbName), runCleaner);
	}

	public DVMetadataRecordHolder(ODataMetadata oDataMetadata, boolean runCleaner, NavigableMap<String, AbstractMetadataRecord> uuidHolder) {
		this(oDataMetadata, runCleaner);
		this.uuidHolder = uuidHolder;
	}

	public boolean isTestMode() {
		return testMode;
	}

	public void setTestMode(boolean testMode) {
		synchronized (lock) {
			this.testMode = testMode;
		}
	}

	public NavigableMap<String, AbstractMetadataRecord> getUuidHolder() {
		return uuidHolder;
	}

	public AbstractMetadataRecord getRecord(String uuid) {
		return uuidHolder.get(uuid);
	}

	public ODataMetadata getODataMetadata() {
		return oDataMetadata;
	}

	public void cleanCache() {
		synchronized (lock) {
			int count = testMode ? THRESHOLD_TO_REMOVE * 2 : 1;
			Set<String> handled = new HashSet<>();
			for (int i = 0; i < count; i++) {
				Map<String, AbstractMetadataRecord> toRemove = new HashMap<>();
				for (Map.Entry<String, AbstractMetadataRecord> entry : uuidHolder.entrySet()) {
					handleRecord(entry.getValue(), handled, toRemove);
				}
				for (AbstractMetadataRecord record : toRemove.values()) {
					removeRecord(record);
				}
			}
		}
	}

	private void handleRecord(AbstractMetadataRecord record, Set<String> handled, Map<String, AbstractMetadataRecord> toRemove) {
		AbstractMetadataRecord parent = record.getParent();
		if (parent != null) {
			boolean contains = parent instanceof ColumnSet ? // we don't store ColumnStore in uuidHolder
					true : uuidHolder.containsKey(parent.getUUID());
			handleRecord(parent, handled, toRemove);
			if (!contains) {
				if (!toRemove.containsKey(record.getUUID())) {
					if (record.getThresholdToRemove() == THRESHOLD_TO_REMOVE) {
						toRemove.put(record.getUUID(), record);
					} else if (!handled.contains(record.getUUID())) {
						record.setThresholdToRemove(record.getThresholdToRemove() + 1);
						handled.add(record.getUUID());
					}
				}
			}
		}
	}

	public AbstractMetadataRecord removeRecord(AbstractMetadataRecord record) {
		if (record != null) {
			if (record instanceof Schema) {
				removeSchemaChildren((Schema) record);
			} else if (record instanceof Table) {
				removeTableChildren((Table) record);
			} else if (record instanceof Procedure) {
				removeProcChildren((Procedure) record);
			}
			uuidHolder.remove(record.getUUID());
			if (test)
				uuidHolder2.remove(record.getUUID());
			record.setUuidChanger(null);
			oDataMetadata.removeRecord(record);
		}
		return record;
	}

	public AbstractMetadataRecord putRecord(AbstractMetadataRecord oldValue, AbstractMetadataRecord newValue) {
		removeRecord(oldValue);
		if (newValue != null) {
			if (newValue instanceof ProcedureParameter)
				return oldValue;
			else if (newValue instanceof Column) {
				ColumnSet columnSet = ((Column) newValue).getParent();
				if (columnSet != null && columnSet.getParent() instanceof Procedure) {
					return oldValue;
				}
			} else if (newValue instanceof Procedure) {
				Procedure proc = (Procedure) newValue;
				proc.getParameters().stream().forEach(p -> putValue(p));
				if (proc.getResultSet() != null && proc.getResultSet().getColumns() != null)
					proc.getResultSet().getColumns().stream().forEach(c -> putValue(c));
			}
			putValue(newValue);
			oDataMetadata.putRecord(newValue);
		}
		return oldValue;
	}

	private void putValue(AbstractMetadataRecord newValue) {
		((ConcurrentSkipListMap)uuidHolder).putIfAbsent(newValue.getUUID(), newValue);
		if (test)
			((ConcurrentSkipListMap)uuidHolder2).putIfAbsent(newValue.getUUID(), newValue);
		newValue.setUuidChanger((oldUUID, newUUID) -> {
			AbstractMetadataRecord record = uuidHolder.remove(oldUUID);
			if (test)
				uuidHolder2.remove(oldUUID);
			if (record != null)
				((ConcurrentSkipListMap)uuidHolder).putIfAbsent(newUUID, record);
			if (test)
				((ConcurrentSkipListMap)uuidHolder2).putIfAbsent(newUUID, record);
		});
		putODataRecord(newValue);
	}

	public void putODataRecord(AbstractMetadataRecord record) {
		oDataMetadata.putRecord(record);
	}

	public ODataSchemaBuilder.ODataSchemaInfo buildODataStructuralMetadata(Schema schema, ODataMetadata.SecurityChecker securityChecker) {
		return oDataMetadata.build(schema, securityChecker);
	}

	public static List initList(Integer initialCapacity) {
		List customList = (List) DVCaller.call(181, initialCapacity);
		return customList != null ? customList : (initialCapacity != null ? new ArrayList<>(initialCapacity) : new ArrayList());
	}

	public static List reInitList(List<? extends AbstractMetadataRecord> oldC, List<? extends AbstractMetadataRecord> newC) {
		List customList = (List) DVCaller.call(182, oldC, newC);
		return customList != null ? customList : newC;
	}

	public static NavigableMap initMap(Comparator<String> comparator) {
		NavigableMap customMap = (NavigableMap) DVCaller.call(183, comparator);
		return customMap != null ? customMap : new ConcurrentSkipListMap<>(comparator);
	}

	public static AbstractMetadataRecord putIntoMap(AbstractMetadataRecord oldValue, AbstractMetadataRecord newValue) {
		return (AbstractMetadataRecord) DVCaller.call(184, oldValue, newValue);
	}

	private void removeTableChildren(Table table) {
		List<Column> cols = table.getColumns();
		if (cols != null) {
			for (Column column : cols) {
				removeRecord(column);
			}
		}
		for (KeyRecord record : table.getAllKeys()) {
			removeRecord(record);
		}
	}

	private void removeProcChildren(Procedure procedure) {
		List<ProcedureParameter> params = procedure.getParameters();
		if (params != null) {
			for (ProcedureParameter param : params) {
				removeRecord(param);
			}
		}
		ColumnSet<Procedure> columnSet = procedure.getResultSet();
		if (columnSet != null && columnSet.getColumns() != null) {
			for (Column column : columnSet.getColumns()) {
				removeRecord(column);
			}
		}
	}

	private void removeSchemaChildren(Schema schema) {
		for (Table table : schema.getTables().values()) {
			removeRecord(table);
		}
		for (Procedure procedure : schema.getProcedures().values()) {
			removeRecord(procedure);
		}
		for (FunctionMethod function : schema.getFunctions().values()) {
			removeRecord(function);
		}
	}

	@Override
	public DVMetadataCache clone(String vdbName, boolean runCleaner) {
		DVMetadataRecordHolder recordHolder = new DVMetadataRecordHolder(this.oDataMetadata.clone(vdbName), runCleaner,
				new ConcurrentSkipListMap(String.CASE_INSENSITIVE_ORDER));
		recordHolder.getUuidHolder().putAll(this.uuidHolder);
		return recordHolder;
	}

	public void refreshCache(MetadataStore metadataStore) {
		synchronized (lock) {
			uuidHolder.clear();
			if (test)
				uuidHolder2.clear();
			if (metadataStore != null) {
				for (Datatype datatype : metadataStore.getDatatypes().values()) {
					uuidHolder.put(datatype.getUUID(), datatype);
				}
				for (Schema schema : metadataStore.getSchemaList()) {
					for (Table table : schema.getTables().values()) {
						if (table.getColumns() != null) {
							for (Column column : table.getColumns()) {
								uuidHolder.put(column.getUUID(), column);
							}
						}
						for (KeyRecord keyRecord : table.getAllKeys()) {
							uuidHolder.put(keyRecord.getUUID(), keyRecord);
						}
						uuidHolder.put(table.getUUID(), table);
					}
					for (Procedure proc : schema.getProcedures().values()) {
						for (ProcedureParameter parameter : proc.getParameters()) {
							uuidHolder.put(parameter.getUUID(), parameter);
						}
						if (proc.getResultSet() != null) {
							for (Column column : proc.getResultSet().getColumns()) {
								uuidHolder.put(column.getUUID(), column);
							}
						}
						uuidHolder.put(proc.getUUID(), proc);
					}
					for (FunctionMethod functionMethod : schema.getFunctions().values()) {
						uuidHolder.put(functionMethod.getUUID(), functionMethod);
					}
					uuidHolder.put(schema.getUUID(), schema);
				}
			}
		}
	}

	public static class UUIDHolderCleaner extends TimerTask {
		private DVMetadataRecordHolder recordHolder;

		public UUIDHolderCleaner(DVMetadataRecordHolder recordHolder) {
			this.recordHolder = recordHolder;
		}

		@Override
		public void run() {
			if (!recordHolder.isTestMode())
				recordHolder.cleanCache();
		}
	}
}
