/*
 * JBoss, Home of Professional Open Source.
 * See the COPYRIGHT.txt file distributed with this work for information
 * regarding copyright ownership.  Some portions may be licensed
 * to Red Hat, Inc. under one or more contributor license agreements.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA.
 * Modified by
 * @author  Data Virtuality GmbH
 * @date    $Date$
 *  
 */
// https://issues.jboss.org/browse/TEIID-2904
package org.teiid.translator.hive;

import static org.teiid.translator.TypeFacility.RUNTIME_NAMES.BIG_INTEGER;
import static org.teiid.translator.TypeFacility.RUNTIME_NAMES.DATE;
import static org.teiid.translator.TypeFacility.RUNTIME_NAMES.DOUBLE;
import static org.teiid.translator.TypeFacility.RUNTIME_NAMES.FLOAT;
import static org.teiid.translator.TypeFacility.RUNTIME_NAMES.INTEGER;
import static org.teiid.translator.TypeFacility.RUNTIME_NAMES.OBJECT;
import static org.teiid.translator.TypeFacility.RUNTIME_NAMES.STRING;
import static org.teiid.translator.TypeFacility.RUNTIME_NAMES.TIMESTAMP;

import java.sql.*;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.teiid.core.types.DataTypeManager;
import org.teiid.core.types.TransformationException;
import org.teiid.core.types.basic.StringToBinaryTypeTransform;
import org.teiid.language.*;
import org.teiid.translator.*;
import org.teiid.translator.jdbc.*;

@Translator(name="hive", description="A translator for hive based database on HDFS")
public class HiveExecutionFactory extends BaseHiveExecutionFactory {
    private static final int MAX_STRING_SIZE = 65534;
    public static String HIVE = "hive"; //$NON-NLS-1$

    public HiveExecutionFactory() {
		setSupportedJoinCriteria(SupportedJoinCriteria.EQUI);
		setForceQuotedIdentifiers(false);
		setSupportsElementaryOlapOperations(true);
        setMaxStringSize(MAX_STRING_SIZE);
        setOversizeStringProcessing(OversizeStringMode.FAIL);
    }
    
    @Override
    public void start() throws TranslatorException {
        super.start();

        convert.addTypeMapping("tinyint", FunctionModifier.BYTE); //$NON-NLS-1$
        convert.addTypeMapping("smallint", FunctionModifier.SHORT); //$NON-NLS-1$
        convert.addTypeMapping("int", FunctionModifier.INTEGER); //$NON-NLS-1$
        convert.addTypeMapping("bigint", FunctionModifier.LONG); //$NON-NLS-1$
        convert.addTypeMapping("decimal(38, 0)", FunctionModifier.BIGINTEGER); //$NON-NLS-1$
        convert.addTypeMapping("boolean", FunctionModifier.BOOLEAN); //$NON-NLS-1$
        convert.addTypeMapping("double", FunctionModifier.DOUBLE); //$NON-NLS-1$
        convert.addTypeMapping("float", FunctionModifier.FLOAT); //$NON-NLS-1$
        convert.addTypeMapping("string", FunctionModifier.STRING); //$NON-NLS-1$
        convert.addTypeMapping("timestamp", FunctionModifier.TIMESTAMP); //$NON-NLS-1$
        convert.addTypeMapping("binary", FunctionModifier.BLOB, FunctionModifier.VARBINARY); //$NON-NLS-1$
        convert.addTypeMapping("decimal", FunctionModifier.BIGDECIMAL); //$NON-NLS-1$
        convert.addTypeMapping("date", FunctionModifier.DATE); //$NON-NLS-1$        
        // unsupported types
        //FunctionModifier.TIME,
        //FunctionModifier.CHAR,
        //FunctionModifier.CLOB,
        //FunctionModifier.XML

        registerFunctionModifier(SourceSystemFunctions.CONVERT, convert);

        registerFunctionModifier(SourceSystemFunctions.BITAND, new AliasModifier("&")); //$NON-NLS-1$
        registerFunctionModifier(SourceSystemFunctions.BITNOT, new AliasModifier("~")); //$NON-NLS-1$
        registerFunctionModifier(SourceSystemFunctions.BITOR, new AliasModifier("|")); //$NON-NLS-1$
        registerFunctionModifier(SourceSystemFunctions.BITXOR, new AliasModifier("^")); //$NON-NLS-1$
        registerFunctionModifier(SourceSystemFunctions.CURDATE, new AliasModifier("unix_timestamp")); //$NON-NLS-1$
        registerFunctionModifier(SourceSystemFunctions.IFNULL, new AliasModifier("coalesce")); //$NON-NLS-1$
        registerFunctionModifier(SourceSystemFunctions.MOD, new ModFunctionModifier("%", getLanguageFactory(), Arrays.asList(TypeFacility.RUNTIME_TYPES.BIG_INTEGER, TypeFacility.RUNTIME_TYPES.BIG_DECIMAL))); //$NON-NLS-1$
        registerFunctionModifier(SourceSystemFunctions.ARRAY_GET, new FunctionModifier() {

            @Override
            public List<?> translate(Function function) {
                return Arrays.asList(function.getParameters().get(0), '[', function.getParameters().get(1), ']');
            }
        });

        addPushDownFunction(HIVE, "lower", STRING, STRING); //$NON-NLS-1$
        addPushDownFunction(HIVE, "upper", STRING, STRING); //$NON-NLS-1$
        addPushDownFunction(HIVE, "negitive", INTEGER, DOUBLE); //$NON-NLS-1$
        addPushDownFunction(HIVE, "negitive", DOUBLE, DOUBLE); //$NON-NLS-1$
        addPushDownFunction(HIVE, "positive", INTEGER, DOUBLE); //$NON-NLS-1$
        addPushDownFunction(HIVE, "positive", DOUBLE, DOUBLE); //$NON-NLS-1$
        addPushDownFunction(HIVE, "ln", DOUBLE, DOUBLE); //$NON-NLS-1$
        addPushDownFunction(HIVE, "reverse", STRING, STRING); //$NON-NLS-1$
        addPushDownFunction(HIVE, "space", INTEGER, STRING); //$NON-NLS-1$
        addPushDownFunction(HIVE, "split", OBJECT, STRING, STRING); //$NON-NLS-1$
        addPushDownFunction(HIVE, "hex", STRING, STRING); //$NON-NLS-1$
        addPushDownFunction(HIVE, "unhex", STRING, STRING); //$NON-NLS-1$
        addPushDownFunction(HIVE, "bin", STRING, BIG_INTEGER); //$NON-NLS-1$
        addPushDownFunction(HIVE, "day", INTEGER, DATE); //$NON-NLS-1$
        addPushDownFunction(HIVE, "datediff", INTEGER, STRING, STRING); //$NON-NLS-1$
        addPushDownFunction(HIVE, "date_add", INTEGER, STRING, INTEGER); //$NON-NLS-1$
        addPushDownFunction(HIVE, "date_sub", INTEGER, STRING, INTEGER); //$NON-NLS-1$
        addPushDownFunction(HIVE, "from_unixtime", STRING, BIG_INTEGER); //$NON-NLS-1$
        addPushDownFunction(HIVE, "from_unixtime", STRING, BIG_INTEGER, STRING); //$NON-NLS-1$
        addPushDownFunction(HIVE, "unix_timestamp", BIG_INTEGER, STRING); //$NON-NLS-1$
        addPushDownFunction(HIVE, "unix_timestamp", BIG_INTEGER, STRING, STRING); //$NON-NLS-1$
        addPushDownFunction(HIVE, "to_date", STRING, STRING); //$NON-NLS-1$
        addPushDownFunction(HIVE, "from_utc_timestamp", TIMESTAMP, TIMESTAMP, STRING); //$NON-NLS-1$
        addPushDownFunction(HIVE, "to_utc_timestamp", TIMESTAMP, TIMESTAMP, STRING); //$NON-NLS-1$
        
        addAggregatePushDownFunction(HIVE, "LEAD", OBJECT, OBJECT); //$NON-NLS-1$
        addAggregatePushDownFunction(HIVE, "LEAD", OBJECT, OBJECT, INTEGER); //$NON-NLS-1$
        addAggregatePushDownFunction(HIVE, "LEAD", OBJECT, OBJECT, INTEGER, OBJECT); //$NON-NLS-1$
        addAggregatePushDownFunction(HIVE, "LAG", OBJECT, OBJECT); //$NON-NLS-1$
        addAggregatePushDownFunction(HIVE, "LAG", OBJECT, OBJECT, INTEGER); //$NON-NLS-1$
        addAggregatePushDownFunction(HIVE, "LAG", OBJECT, OBJECT, INTEGER, OBJECT); //$NON-NLS-1$
        addAggregatePushDownFunction(HIVE, "FIRST_VALUE", OBJECT, OBJECT); //$NON-NLS-1$
        addAggregatePushDownFunction(HIVE, "LAST_VALUE", OBJECT, OBJECT); //$NON-NLS-1$
        addAggregatePushDownFunction(HIVE, "PERCENT_RANK", FLOAT); //$NON-NLS-1$
        addAggregatePushDownFunction(HIVE, "CUME_DIST", FLOAT); //$NON-NLS-1$
        addAggregatePushDownFunction(HIVE, "NTILE", BIG_INTEGER, INTEGER); //$NON-NLS-1$
    }    
    
    @Override
    public List<String> getSupportedFunctions() {
        List<String> supportedFunctions = new ArrayList<>(super.getSupportedFunctions());

        supportedFunctions.add(SourceSystemFunctions.ABS);
        supportedFunctions.add(SourceSystemFunctions.ACOS);
        supportedFunctions.add(SourceSystemFunctions.ARRAY_GET);
        supportedFunctions.add(SourceSystemFunctions.ASIN);
        supportedFunctions.add(SourceSystemFunctions.ASCII);
        supportedFunctions.add(SourceSystemFunctions.ATAN);
        supportedFunctions.add(SourceSystemFunctions.BITAND);
        supportedFunctions.add(SourceSystemFunctions.BITNOT);
        supportedFunctions.add(SourceSystemFunctions.BITOR);
        supportedFunctions.add(SourceSystemFunctions.BITXOR);
        supportedFunctions.add(SourceSystemFunctions.CEILING);
        supportedFunctions.add(SourceSystemFunctions.COALESCE);
        supportedFunctions.add(SourceSystemFunctions.CONCAT);
        supportedFunctions.add(SourceSystemFunctions.COS);
        supportedFunctions.add(SourceSystemFunctions.CONVERT);
        supportedFunctions.add(SourceSystemFunctions.CURDATE);
        supportedFunctions.add(SourceSystemFunctions.CURTIME);
        supportedFunctions.add(SourceSystemFunctions.DEGREES);
        supportedFunctions.add(SourceSystemFunctions.DAYOFMONTH);
        supportedFunctions.add(SourceSystemFunctions.EXP);
        supportedFunctions.add(SourceSystemFunctions.FLOOR);
        supportedFunctions.add(SourceSystemFunctions.HOUR);
        supportedFunctions.add(SourceSystemFunctions.IFNULL);
        supportedFunctions.add(SourceSystemFunctions.LCASE);
        supportedFunctions.add(SourceSystemFunctions.LOCATE);
        supportedFunctions.add(SourceSystemFunctions.LPAD);
        supportedFunctions.add(SourceSystemFunctions.LENGTH);
        supportedFunctions.add(SourceSystemFunctions.LTRIM);
        supportedFunctions.add(SourceSystemFunctions.LOG);
        supportedFunctions.add(SourceSystemFunctions.LOG10);
        supportedFunctions.add(SourceSystemFunctions.MINUTE);
        supportedFunctions.add(SourceSystemFunctions.MOD);
        supportedFunctions.add(SourceSystemFunctions.POWER);
        supportedFunctions.add(SourceSystemFunctions.SECOND);
        supportedFunctions.add(SourceSystemFunctions.SQRT);
        supportedFunctions.add(SourceSystemFunctions.RADIANS);
        supportedFunctions.add(SourceSystemFunctions.ROUND);
        supportedFunctions.add(SourceSystemFunctions.RTRIM);
        supportedFunctions.add(SourceSystemFunctions.RPAD);
        supportedFunctions.add(SourceSystemFunctions.MONTH);
        supportedFunctions.add(SourceSystemFunctions.PI);
        supportedFunctions.add(SourceSystemFunctions.SIN);
        supportedFunctions.add(SourceSystemFunctions.SUBSTRING);
        supportedFunctions.add(SourceSystemFunctions.TAN);
        supportedFunctions.add(SourceSystemFunctions.TRIM);
        supportedFunctions.add(SourceSystemFunctions.UCASE);
        supportedFunctions.add(SourceSystemFunctions.YEAR);
        supportedFunctions.add(SourceSystemFunctions.RAND);
        return supportedFunctions;
    }    
    
    @Override
    public boolean supportsAggregatesEnhancedNumeric() {
        return true;
    }
    
    @Override
    public boolean supportsElementaryOlapOperations() {
        return true;
    }    
    
    @Override
    public boolean useParensForJoins() {
    	return true;
    }

    @Override
    public List<?> translate(LanguageObject obj, ExecutionContext context, boolean isUpdate) {
	    if (obj instanceof Comparison) {
            Comparison obj1 = (Comparison) obj;
            if (isNumericExpression(obj1.getLeftExpression())) {
                return Arrays.asList("cast((", obj1.getLeftExpression(), ") as decimal)", obj1.getOperator().toString(),
                        "cast((", obj1.getRightExpression(), ") as decimal)");
            }
        }
        if (obj instanceof In) {
            In obj1 = (In) obj;
            if (isNumericExpression(obj1.getLeftExpression())) {
                List<Object> list = new ArrayList<>();
                list.add("cast((");
                list.add(obj1.getLeftExpression());
                list.add(") as decimal) IN (");
                boolean comma = false;
                for (Expression e : obj1.getRightExpressions()) {
                    if (comma)
                        list.add(",");
                    else
                        comma = true;
                    list.add(e);
                }
                list.add(")");
                return list;
            }
        }
        return super.translate(obj, context, isUpdate);
    }

    private boolean isNumericExpression(Expression e) {
        final Class<?> t = e.getType();
        return t.equals(Integer.class)
                || t.equals(Long.class)
                || t.equals(Short.class)
                || t.equals(Byte.class);
    }

    @Override
    public String translateLiteralTimestamp(Timestamp timestampValue) {
        return "'" + formatDateValue(timestampValue) + "'";
    }

    @Override
    public void bindValue(PreparedStatement stmt, Object param, Class<?> paramType, int i) throws SQLException {
        if (Clob.class.isAssignableFrom(paramType)) {
            Clob clob = (Clob) param;
            stmt.setString(i, clob.getSubString(1, (int) clob.length()));
            return;
        }
        if (paramType.equals(Character.class) || paramType.equals(String.class)) {
            stmt.setString(i, param.toString());
            return;
        }
        super.bindValue(stmt, param, paramType, i);
    }

    @Override
    public JDBCMetadataProcessor.ITypeConverter getTypeConverter() {
        return new JDBCMetadataProcessor.ITypeConverter() {
            @Override
            public String getRuntimeType(int type, String typeName, int precision, int scale) {
                if ("varchar".equals(typeName) || "string".equals(typeName)) {
                    if (precision > 0 && precision <= MAX_STRING_SIZE)
                        return DataTypeManager.DefaultDataTypes.STRING;
                    return DataTypeManager.DefaultDataTypes.CLOB;
                }
                if ("binary".equals(typeName)) {
                    return DataTypeManager.DefaultDataTypes.BLOB;
                }
                return null;
            }
        };
    }

    public Object retrieveValue(ResultSet results, int columnIndex, Class<?> expectedType) throws SQLException {
        if (TypeFacility.RUNTIME_TYPES.VARBINARY.equals(expectedType)) {
            try {
                return new StringToBinaryTypeTransform().transformDirect(results.getString(columnIndex));
            } catch (TransformationException e) {
                throw new SQLException(e);
            }
        }
        return super.retrieveValue(results, columnIndex, expectedType);
    }

    @Override
    public Object retrieveValue(CallableStatement results, int parameterIndex, Class<?> expectedType) throws SQLException {
        if (TypeFacility.RUNTIME_TYPES.VARBINARY.equals(expectedType)) {
            try {
                return new StringToBinaryTypeTransform().transformDirect(results.getString(parameterIndex));
            } catch (TransformationException e) {
                throw new SQLException(e);
            }
        }
        return super.retrieveValue(results, parameterIndex, expectedType);
    }
}
