/*
 * JBoss, Home of Professional Open Source.
 * See the COPYRIGHT.txt file distributed with this work for information
 * regarding copyright ownership.  Some portions may be licensed
 * to Red Hat, Inc. under one or more contributor license agreements.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA.
 * Modified by
 * @author  Data Virtuality GmbH
 * @date    $Date$
 *  
 */
// https://issues.jboss.org/browse/TEIID-2904
// https://issues.jboss.org/browse/TEIID-2281
package org.teiid.translator.hive;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.datavirtuality.dv.core.interfaces.DVCaller;
import org.apache.commons.lang3.tuple.ImmutablePair;
import org.apache.commons.lang3.tuple.Pair;
import org.teiid.metadata.Column;
import org.teiid.metadata.MetadataFactory;
import org.teiid.metadata.Table;
import org.teiid.translator.MetadataProcessor;
import org.teiid.translator.TranslatorException;
import org.teiid.translator.TranslatorProperty;
import org.teiid.translator.TranslatorProperty.PropertyType;
import org.teiid.translator.TypeFacility;
import org.teiid.translator.jdbc.JDBCMetadataProcessor;

public class HiveMetadataProcessor extends JDBCMetadataProcessor implements MetadataProcessor<Connection>{
	
	private boolean trimColumnNames;
	private boolean useDatabaseMetaData;

	public HiveMetadataProcessor() {
        setImportKeys(false);
		setUseCatalogName(false);
	}
	
	@Override
    public void process(MetadataFactory metadataFactory, Connection conn)    throws TranslatorException {
        try {
            getConnectorMetadata(conn, metadataFactory, new HiveExecutionFactory().getTypeConverter());
        } catch (SQLException e) {
            throw new TranslatorException(e);
        }
    }
	
    @Override
    public void getConnectorMetadata(Connection conn, MetadataFactory metadataFactory, ITypeConverter typeConv)
            throws SQLException, TranslatorException {
		this.typeConv = typeConv;
        if (useDatabaseMetaData) {
        	setLoadMetadataWithJdbc(true);
            super.getConnectorMetadata(conn, metadataFactory, typeConv);
            return;
        }
        List<Pair<String, String>> tables = getTables(conn);
        for (Pair<String, String> table:tables) {
//            if (shouldExclude(table)) {
//                continue;
//            }
            addTable(table.getLeft(), table.getRight(), conn, metadataFactory);
        }
    }
    
    private List<Pair<String, String>> getTables(Connection conn) throws SQLException {
    	List<String> databases = new ArrayList<>();
		//noinspection unchecked
		Collection<String> schemaRegexps = (Collection<String>) DVCaller.call(179, getSchemaPattern());
		if (schemaRegexps == null) {
			schemaRegexps = new HashSet<>();
			schemaRegexps.add(null);
		}
		Collection<Pattern> schemaPats = new HashSet<>(schemaRegexps.size());
		for (String schemaRe : schemaRegexps) {
			if (schemaRe != null)
				schemaPats.add(Pattern.compile(schemaRe));
			else
				schemaPats.add(null);
		}
    	try (Statement stmt = conn.createStatement()) {
			try (ResultSet rs = stmt.executeQuery("SHOW DATABASES")) { //$NON-NLS-1$
				while (rs.next()) {
					final String schema = rs.getString(1);
					for (Pattern schemaPat : schemaPats) {
						if (schemaPat == null || schemaPat.matcher(schema).matches()) {
							databases.add(schema);
							break;
						}
					}
				}
			}
		}
		String tableRe = (String) DVCaller.call(180, getTableNamePattern());
		Pattern tablePat = null;
		if (tableRe != null)
			tablePat = Pattern.compile(tableRe);
    	ArrayList<Pair<String, String>> tables = new ArrayList<>();
		try (Statement stmt = conn.createStatement()) {
			for (String database : databases) {
				try (ResultSet rs = stmt.executeQuery("SHOW TABLES in " + database)) { //$NON-NLS-1$
					while (rs.next()) {
						final String table = rs.getString(1);
						if (tablePat == null || tablePat.matcher(table).matches()) {
							tables.add(ImmutablePair.of(database, table));
						}
					}
				}
			}
		}
		return tables;
	}
	
	private String getRuntimeType(String type, int precision, int scale) {
		if (typeConv != null) {
			String runtimeType = typeConv.tuneRuntimeType(0, type, precision, scale);
			if (runtimeType != null)
				return runtimeType;
		}
		type = type.toLowerCase();
		switch (type) {
			case "int":
				return TypeFacility.RUNTIME_NAMES.INTEGER;
			case "tinyint":
				return TypeFacility.RUNTIME_NAMES.BYTE;
			case "smallint":
				return TypeFacility.RUNTIME_NAMES.SHORT;
			case "bigint":
				return TypeFacility.RUNTIME_NAMES.LONG;
			case "float":
				return TypeFacility.RUNTIME_NAMES.FLOAT;
			case "double":
				return TypeFacility.RUNTIME_NAMES.DOUBLE;
			case "boolean":
				return TypeFacility.RUNTIME_NAMES.BOOLEAN;
			case "decimal":
				if (scale == 0)
					return TypeFacility.RUNTIME_NAMES.BIG_INTEGER;
				return TypeFacility.RUNTIME_NAMES.BIG_DECIMAL;
			case "timestamp":
				return TypeFacility.RUNTIME_NAMES.TIMESTAMP;
			case "date":
				return TypeFacility.RUNTIME_NAMES.DATE;
			case "binary":
				return TypeFacility.RUNTIME_NAMES.VARBINARY;
			case "varchar":
			case "string":
			default:
				return TypeFacility.RUNTIME_NAMES.STRING;
		}
 	}

	private int fixPrecision(String type, int precision, int scale) {
		type = type.toLowerCase();
		switch (type) {
			case "int":
				return 10;
			case "tinyint":
				return 3;
			case "smallint":
				return 5;
			case "bigint":
				return 19;
			case "float":
				return 7;
			case "double":
				return 15;
		}
		return precision;
	}

	private final Pattern patTypesWithPrecisionAndScale = Pattern.compile("^(.+)\\((\\d+),(\\d+)\\)$");
	private final Pattern patTypesWithLength = Pattern.compile("^(.+)\\((\\d+)\\)$");
	private void addTable(String database, String tableName, Connection conn, MetadataFactory metadataFactory) throws SQLException {
		String fullName = database + "." + tableName;
		Table table = addTable(metadataFactory, null, database, tableName, null, fullName, null);
		if (table == null) {
			return;
		}
		Statement stmt = conn.createStatement();
		ResultSet rs =  stmt.executeQuery("DESCRIBE "+fullName); //$NON-NLS-1$
		while (rs.next()){
			String name = rs.getString(1);
			if (this.trimColumnNames) {
                name = name.trim();
            }
			// After the last column for a table Hive sends an empty row and them gives some '# Partition Information' which we don't use, so we just quit.
			if ("".equals(name))
				break;
			String type = rs.getString(2);
			String nativeType = type;
			String runtimeType;
			int precision = 0;
			int scale = 0;
			if (type != null) {
				type = type.trim();
				final Matcher m1 = patTypesWithPrecisionAndScale.matcher(type);
				final Matcher m2 = patTypesWithLength.matcher(type);
				if (m1.matches()) {
					type = m1.group(1);
					precision = Integer.parseInt(m1.group(2));
					scale = Integer.parseInt(m1.group(3));
				} else if (m2.matches()) {
					type = m2.group(1);
					precision = Integer.parseInt(m2.group(2));
				}
				runtimeType = getRuntimeType(type, precision, scale);
				precision = fixPrecision(type, precision, scale);
			} else
				runtimeType = TypeFacility.RUNTIME_NAMES.STRING;
			if (typeConv != null)
				if (!typeConv.filterColumn(name, 0, type, precision, scale, precision))
					continue;
			Column column = metadataFactory.addColumn(name, runtimeType, table);
			column.setLength(precision);
			column.setPrecision(precision);
			column.setScale(scale);
			column.setNameInSource(name);
			column.setUpdatable(true);
			column.setNativeType(nativeType);
			column.setNativeLength(precision);
			column.setNativePrecision(precision);
			column.setNativeScale(scale);
			if (typeConv != null)
				typeConv.tuneColumnType(column, 0, type, precision, scale);
		}
		rs.close();
	}

	public void setTrimColumnNames(boolean trimColumnNames) {
        this.trimColumnNames = trimColumnNames;
    }

    @TranslatorProperty(display="Trim Columns", category=PropertyType.IMPORT, description="Trim column names read from the database")
    public boolean isTrimColumnNames() {
        return trimColumnNames;
    }
    
	@TranslatorProperty(display="Use DatabaseMetaData", category=PropertyType.IMPORT, description= "Use DatabaseMetaData (typical JDBC logic) for importing")
    public boolean isUseDatabaseMetaData() {
        return useDatabaseMetaData;
    }

    public void setUseDatabaseMetaData(boolean useDatabaseMetaData) {
        this.useDatabaseMetaData = useDatabaseMetaData;
    }
}
