package org.teiid.translator.hive;

import com.datavirtuality.dv.core.interfaces.DVCaller;
import org.apache.commons.lang3.StringUtils;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.hive.common.type.HiveChar;
import org.apache.hadoop.hive.ql.io.parquet.serde.ArrayWritableObjectInspector;
import org.apache.hadoop.hive.ql.io.parquet.serde.ParquetHiveSerDe;
import org.apache.hadoop.hive.ql.io.parquet.write.DataWritableWriteSupport;
import org.apache.hadoop.hive.serde.serdeConstants;
import org.apache.hadoop.hive.serde2.SerDeException;
import org.apache.hadoop.hive.serde2.io.DateWritable;
import org.apache.hadoop.hive.serde2.io.HiveCharWritable;
import org.apache.hadoop.hive.serde2.io.ParquetHiveRecord;
import org.apache.hadoop.hive.serde2.io.TimestampWritable;
import org.apache.hadoop.hive.serde2.typeinfo.StructTypeInfo;
import org.apache.hadoop.hive.serde2.typeinfo.TypeInfo;
import org.apache.hadoop.hive.serde2.typeinfo.TypeInfoUtils;
import org.apache.hadoop.io.*;
import org.teiid.core.TeiidRuntimeException;
import org.teiid.core.types.BinaryType;
import org.teiid.core.types.DataTypeManager;
import org.teiid.core.types.TransformationException;
import org.teiid.language.*;
import org.teiid.metadata.BaseColumn.NullType;
import org.teiid.metadata.Column;
import org.teiid.metadata.RuntimeMetadata;
import org.teiid.query.sql.symbol.TextLine;
import org.teiid.translator.ExecutionContext;
import org.teiid.translator.TranslatorException;
import org.teiid.translator.jdbc.JDBCUpdateExecution;
import parquet.column.ParquetProperties.WriterVersion;
import parquet.hadoop.ParquetWriter;
import parquet.hadoop.metadata.CompressionCodecName;
import parquet.schema.MessageType;
import parquet.schema.MessageTypeParser;

import java.io.BufferedOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.net.URI;
import java.nio.charset.StandardCharsets;
import java.sql.Date;
import java.sql.*;
import java.util.*;

public class HiveUpdateExecution extends JDBCUpdateExecution {
	private BaseHiveExecutionFactory hiveExecutionFactory;

	public HiveUpdateExecution(Command command, ExecutionContext executionContext, RuntimeMetadata metadata, Connection conn, BaseHiveExecutionFactory hiveExecutionFactory) {
		super(command, conn, executionContext, hiveExecutionFactory);
		this.hiveExecutionFactory = hiveExecutionFactory;
	}

	@Override
	public void execute() throws TranslatorException {
		Insert insert = (Insert) command;
		Iterator<? extends List<?>> it;
		if (insert.getParameterValues() != null) {
			if (!insert.getParameterValues().hasNext()) {
				result = new int[] { 0 };
				return;
			}
			it = insert.getParameterValues();
		} else
			it = Collections.singletonList(((ExpressionValueSource) insert.getValueSource()).getValues()).iterator();
		try {
			DVCaller.callWithExc(76, command, connection);
		} catch (Throwable err) {
			throw new TranslatorException(err);
		}
		uploadToHdfs(insert.getTable().getMetadataObject().getNameInSource(), insert.getColumns(), it);
	}

	private void uploadToHdfs(String tableName, List<ColumnReference> columns, Iterator<? extends List<?>> parameterValues) throws TranslatorException {
        try {
    		Configuration conf = new Configuration();
	        String endpoint = hiveExecutionFactory.webhdfsEndpoint();
			conf.set("fs.defaultFS", endpoint);
    		conf.set("fs.webhdfs.impl", org.apache.hadoop.hdfs.web.WebHdfsFileSystem.class.getName());
			conf.set("fs.file.impl", "com.globalmentor.apache.hadoop.fs.BareLocalFileSystem");
    		FileSystem fs = FileSystem.get(URI.create(endpoint), conf);
			String fileName = "dv_table_data_" + UUID.randomUUID().toString().replace("-", "") + ".db";
			String fileWithPath = hiveExecutionFactory.webhdfsLocation() + "/" + fileName;
			Path temp = new Path(endpoint + "/" + fileWithPath);
			int count = 0;
			try {
				switch (hiveExecutionFactory.tableType()) {
				case PARQUET:
					count = writeParquet(conf, temp, columns, parameterValues);
					break;
				default:
					count = writeTextfile(fs, temp, columns, parameterValues);
					break;
				}
				String sql = "LOAD DATA INPATH '" + fileWithPath + "' INTO TABLE " + tableName;
				executeSql(sql);
				result = new int[] { count };
			} finally {
				if (fs != null) {
					fs.close();
				}
			}
		} catch (Exception e) {
			throw new TranslatorException(e);
		}
	}

	private int executeSql(String sql) throws SQLException {
		Statement stm = null;
		try {
			stm = connection.createStatement();
			return stm.executeUpdate(sql);
		} finally {
			if (stm != null) {
				try {
					stm.close();
				} catch (SQLException e) {
					stm = null;
				}
			}
		}
	}

	private int writeTextfile(FileSystem fs, Path temp, List<ColumnReference> columns, Iterator<? extends List<?>> parameterValues) throws IOException, TransformationException {
		int count = 0;
		try (OutputStream out = new BufferedOutputStream(fs.create(temp, true))) {
			while (parameterValues.hasNext()) {
				//noinspection rawtypes
				final List row = parameterValues.next();
				for (int i = 0; i < row.size(); i++) {
					Object obj = row.get(i);
					Object obj1 = prepareValue(obj);
					if (obj != obj1)
						//noinspection unchecked
						row.set(i, obj1);
				}
				out.write(convertListToString(generateValueString(columns, row)).getBytes(StandardCharsets.UTF_8));
				count++;
			}
		}
		return count;
	}

	private String convertListToString(String[] list) {
		StringBuilder sb = new StringBuilder();
		for (Object obj : list) {
			sb.append(obj.toString());
		}
		return sb.toString();
	}

	private String[] generateValueString(List<ColumnReference> columns, List<?> values) throws TransformationException {
		@SuppressWarnings("unchecked")
		List<Object> values2 = (List<Object>) values;
		TextLine textLine = new TextLine();
		textLine.setDelimiter('|');
		textLine.setQuote('\0');
		return TextLine.evaluate(values2, new TextLine.ValueExtractor<Object>() {
			public Object getValue(Object t) {
				if (t == null)
					return "\\N";
				if (t instanceof Time)
					return new Timestamp(((Time) t).getTime());
				if (t instanceof Blob)
					throw new TeiidRuntimeException("Can't INSERT a BLOB column using TEXTFILE. Use tableType=PARQUET translator property for the datasource to create parquet tables (but note that you can't properly modify existing ones that way).");
				return t;
			}
		}, textLine);
	}

	private Object prepareValue(Object value) {
		if (value instanceof Literal)
			value = ((Literal) value).getValue();
		return value;
	}

	private int writeParquet(Configuration config, Path out, final List<ColumnReference> columns, Iterator<? extends List<?>> parameterValues) throws IOException, SQLException, SerDeException {
		String schemaText = "message temp { " + generateParquetSchema(columns) + " }";
		final MessageType schema = MessageTypeParser.parseMessageType(schemaText);
		DataWritableWriteSupport.setSchema(schema, config);
		config.set(DataWritableWriteSupport.PARQUET_HIVE_SCHEMA, schemaText);

		@SuppressWarnings({ "unchecked", "rawtypes" })
		ParquetWriter<ParquetHiveRecord> writer = new ParquetWriter(out, new DataWritableWriteSupport(), CompressionCodecName.GZIP,
				134217728, 1048576, 1048576, true, false, WriterVersion.PARQUET_2_0, config);
		Properties tbl = new Properties();
		StructTypeInfo rowTypeInfo = new StructTypeInfo();
		buildParquetCols(columns, tbl, rowTypeInfo);
		ParquetHiveSerDe ser = new ParquetHiveSerDe();
		ArrayWritableObjectInspector ins = new ArrayWritableObjectInspector(rowTypeInfo);
		ser.initialize(config, tbl);
		int count = 0;
		while (parameterValues.hasNext()) {
			List<?> row = parameterValues.next();
			int size = row.size();
			Writable[] values = new Writable[size];
			for (int i = 0; i < size; i++)
				values[i] = generateParquetValue(columns.get(i), row.get(i));
			ParquetHiveRecord value = (ParquetHiveRecord) ser.serialize(new ArrayWritable(Writable.class, values), ins);
			writer.write(value);
			count++;
		}
		writer.close();
		return count;
	}

	public void buildParquetCols(final List<ColumnReference> columns, Properties tbl, StructTypeInfo rowTypeInfo) {
		ArrayList<String> lNames = new ArrayList<String>();
		ArrayList<String> lTypeNames = new ArrayList<String>();
		ArrayList<TypeInfo> lTypes = new ArrayList<TypeInfo>();
		for (ColumnReference col : columns) {
			lNames.add(col.getName());
			String type = convertColType(col);
			lTypeNames.add(type);
			lTypes.add(TypeInfoUtils.getTypeInfoFromTypeString(type));
		}
		rowTypeInfo.setAllStructFieldNames(lNames);
		rowTypeInfo.setAllStructFieldTypeInfos(lTypes);
		tbl.setProperty(serdeConstants.LIST_COLUMNS, StringUtils.join(lNames, ','));
		tbl.setProperty(serdeConstants.LIST_COLUMN_TYPES, StringUtils.join(lTypeNames, ','));
	}

	public String convertColType(ColumnReference col) {
		String type = col.getMetadataObject().getDatatype().getRuntimeTypeName();
		switch (type) {
		case DataTypeManager.DefaultDataTypes.BYTE:
			return "tinyint";
		case DataTypeManager.DefaultDataTypes.SHORT:
			return "smallint";
		case DataTypeManager.DefaultDataTypes.INTEGER:
			return "int";
		case DataTypeManager.DefaultDataTypes.LONG:
			return "bigint";
		case DataTypeManager.DefaultDataTypes.BIG_DECIMAL:
			return "binary";
		case DataTypeManager.DefaultDataTypes.BIG_INTEGER:
			return "binary";
		case DataTypeManager.DefaultDataTypes.BLOB:
			return "binary";
		case DataTypeManager.DefaultDataTypes.VARBINARY:
			return "binary";
		case DataTypeManager.DefaultDataTypes.CLOB:
			return "string";
		case DataTypeManager.DefaultDataTypes.TIME:
			return "timestamp";
		case DataTypeManager.DefaultDataTypes.CHAR:
			return "char(" + col.getMetadataObject().getLength() + ")";
		}
		return type;
	}

	private String generateParquetSchema(List<ColumnReference> columns) {
		StringBuilder result = new StringBuilder();
		for (ColumnReference ref : columns) {
			result.append(' ');
			final Column col = ref.getMetadataObject();
			if (col.getDatatype().getNullType() == NullType.No_Nulls)
				result.append("required");
			else
				result.append("optional");
			result.append(' ');
			// [INT64, INT32, BOOLEAN, BINARY, FLOAT, DOUBLE, INT96, FIXED_LEN_BYTE_ARRAY]
			String prefix = null;
			String suffix = null;
			String type = col.getDatatype().getRuntimeTypeName();
			switch (type) {
			case DataTypeManager.DefaultDataTypes.INTEGER:
				prefix = "int32";
				suffix = "INT_32";
				break;
			case DataTypeManager.DefaultDataTypes.BOOLEAN:
				prefix = "boolean";
				break;
			case DataTypeManager.DefaultDataTypes.LONG:
				prefix = "int64";
				suffix = "INT_64";
				break;
			case DataTypeManager.DefaultDataTypes.BYTE:
				prefix = "int32";
				suffix = "INT_8";
				break;
			case DataTypeManager.DefaultDataTypes.SHORT:
				prefix = "int32";
				suffix = "INT_16";
				break;
			case DataTypeManager.DefaultDataTypes.FLOAT:
				prefix = "float";
				break;
			case DataTypeManager.DefaultDataTypes.DOUBLE:
				prefix = "double";
				break;
			case DataTypeManager.DefaultDataTypes.CLOB:
			case DataTypeManager.DefaultDataTypes.STRING:
			case DataTypeManager.DefaultDataTypes.CHAR:
				prefix = "binary";
				suffix = "UTF8";
				break;
			case DataTypeManager.DefaultDataTypes.BIG_INTEGER:
			case DataTypeManager.DefaultDataTypes.BIG_DECIMAL:
				prefix = "binary";
				suffix = String.format("DECIMAL(%d,%d)", col.getPrecision(), col.getScale());
				break;
			case DataTypeManager.DefaultDataTypes.DATE:
				prefix = "int32";
				suffix = "DATE";
				break;
			case DataTypeManager.DefaultDataTypes.TIME:
			case DataTypeManager.DefaultDataTypes.TIMESTAMP:
				prefix = "int96";
				break;
			default:
				prefix = "binary";
			}
			result.append(prefix).append(' ').append(ref.getName());
			if (suffix != null)
				result.append(" (").append(suffix).append(')');
			result.append(';');
		}
		return result.toString();
	}

	private Writable generateParquetValue(ColumnReference ref, Object obj) throws SQLException {
		obj = prepareValue(obj);
		if (obj == null)
			return null;
		final Column col = ref.getMetadataObject();
		String type = col.getDatatype().getRuntimeTypeName();
		switch (type) {
		case DataTypeManager.DefaultDataTypes.INTEGER:
			return new IntWritable((Integer) obj);
		case DataTypeManager.DefaultDataTypes.BOOLEAN:
			return new BooleanWritable((Boolean) obj);
		case DataTypeManager.DefaultDataTypes.LONG:
			return new LongWritable((Long) obj);
		case DataTypeManager.DefaultDataTypes.BYTE:
			return new IntWritable((Byte) obj);
		case DataTypeManager.DefaultDataTypes.SHORT:
			return new IntWritable((Short) obj);
		case DataTypeManager.DefaultDataTypes.FLOAT:
			return new FloatWritable((Float) obj);
		case DataTypeManager.DefaultDataTypes.DOUBLE:
			return new DoubleWritable((Double) obj);
		case DataTypeManager.DefaultDataTypes.DATE:
			return new DateWritable(new Date(convertToUTC((Date) obj)));
		case DataTypeManager.DefaultDataTypes.TIME:
		case DataTypeManager.DefaultDataTypes.TIMESTAMP:
			return new TimestampWritable(new Timestamp(convertToUTC((java.util.Date) obj)));
		case DataTypeManager.DefaultDataTypes.BLOB:
			if (obj instanceof Blob) {
				Blob blob = (Blob) obj;
				return new BytesWritable(blob.getBytes(1, (int) blob.length()));
			} else
				return new BytesWritable(((BinaryType) obj).getBytes());
		case DataTypeManager.DefaultDataTypes.STRING:
			return new BytesWritable((obj.toString()).getBytes(StandardCharsets.UTF_8));
		case DataTypeManager.DefaultDataTypes.CHAR:
			return new HiveCharWritable(new HiveChar(((Character) obj).toString(), col.getLength()));
		case DataTypeManager.DefaultDataTypes.CLOB:
			Clob clob = (Clob) obj;
			return new BytesWritable((clob.getSubString(1, (int) clob.length())).getBytes(StandardCharsets.UTF_8));
		case DataTypeManager.DefaultDataTypes.BIG_INTEGER:
			return new BytesWritable(((BigInteger) obj).toByteArray());
		case DataTypeManager.DefaultDataTypes.BIG_DECIMAL:
			return new BytesWritable(((BigDecimal) obj).setScale(col.getScale(), BigDecimal.ROUND_FLOOR).unscaledValue().toByteArray());
		}
		return new BytesWritable(obj.toString().getBytes());
	}

	private long convertToUTC(java.util.Date value) {
    	Calendar cal = GregorianCalendar.getInstance();
    	cal.setTimeInMillis(value.getTime());
    	int offset = TimeZone.getDefault().getOffset(cal.getTime().getTime());
    	cal.add(Calendar.MILLISECOND, offset);
    	return cal.getTime().getTime();
    }
}
