/*
 * JBoss, Home of Professional Open Source.
 * See the COPYRIGHT.txt file distributed with this work for information
 * regarding copyright ownership.  Some portions may be licensed
 * to Red Hat, Inc. under one or more contributor license agreements.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA.
 * Modified by
 * @author  Data Virtuality GmbH
 * @date    $Date$
 *  
 */
// https://issues.jboss.org/browse/TEIID-4631
package org.teiid.translator.jdbc.teradata;

import com.datavirtuality.dv.core.interfaces.DVCaller;
import org.apache.commons.io.IOUtils;
import org.teiid.core.types.BinaryType;
import org.teiid.core.types.DataTypeManager;
import org.teiid.language.*;
import org.teiid.metadata.Column;
import org.teiid.translator.*;
import org.teiid.translator.jdbc.*;
import org.teiid.translator.jdbc.JDBCMetadataProcessor.ITypeConverter;

import java.io.IOException;
import java.sql.Date;
import java.sql.*;
import java.util.*;

import static org.teiid.translator.TypeFacility.RUNTIME_NAMES.*;



/** 
 * Teradata database Release 12
 */
@Translator(name="teradata", description="A translator for Teradata Database")
public class TeradataExecutionFactory extends JDBCExecutionFactory {
	public static String TERADATA = "teradata"; //$NON-NLS-1$
	private static final int NUMBER_PREC = 255;
	private static final int NUMBER_SCALE = 130;

	protected ConvertModifier convert = new ConvertModifier();
	
	private boolean supportsRandFunction = false;
	@TranslatorProperty(display="Supports RAND function", description="Support RAND function",advanced=true)
	public boolean supportsRandFunction() {
		return this.supportsRandFunction;
	}

	@SuppressWarnings("unused")
	public void setSupportsRandFunction(boolean supportsRandFunction) {
		this.supportsRandFunction = supportsRandFunction;
	}
	
    public TeradataExecutionFactory() {
    	setSupportsOuterJoins(true);
		setSupportsInnerJoins(true);
		setSupportsFullOuterJoins(true);
		setSupportsMd5Function(true);
		setTransactionSupport(TransactionSupport.LOCAL);
		//noinspection deprecation
		setSupportsNativeQueries(true);
		setMaxStringSize(32000);
    }

	@Override
	public void start() throws TranslatorException {
		super.start();
		convert.addTypeMapping("byteint", FunctionModifier.BYTE, FunctionModifier.SHORT, FunctionModifier.BOOLEAN); //$NON-NLS-1$
		convert.addTypeMapping("double precision", FunctionModifier.DOUBLE); //$NON-NLS-1$
		convert.addTypeMapping("numeric(38,0)", FunctionModifier.BIGINTEGER); //$NON-NLS-1$
		convert.addTypeMapping("char(1)", FunctionModifier.CHAR); //$NON-NLS-1$

    	convert.addConvert(FunctionModifier.TIMESTAMP, FunctionModifier.TIME, new CastModifier("TIME")); //$NON-NLS-1$
    	convert.addConvert(FunctionModifier.TIMESTAMP, FunctionModifier.DATE,  new CastModifier("DATE")); //$NON-NLS-1$ 
    	convert.addConvert(FunctionModifier.TIME, FunctionModifier.TIMESTAMP, new CastModifier("TIMESTAMP")); //$NON-NLS-1$
    	convert.addConvert(FunctionModifier.DATE, FunctionModifier.TIMESTAMP,  new CastModifier("TIMESTAMP")); //$NON-NLS-1$

    	convert.addConvert(FunctionModifier.STRING, FunctionModifier.INTEGER, new CastModifier("integer")); //$NON-NLS-1$
    	convert.addConvert(FunctionModifier.STRING, FunctionModifier.BIGDECIMAL, new CastModifier("decimal(37,5)"));//$NON-NLS-1$
    	convert.addConvert(FunctionModifier.STRING, FunctionModifier.BIGINTEGER, new CastModifier("numeric(38,0)"));//$NON-NLS-1$
    	convert.addConvert(FunctionModifier.STRING, FunctionModifier.FLOAT, new CastModifier("float"));//$NON-NLS-1$
    	convert.addConvert(FunctionModifier.STRING, FunctionModifier.BOOLEAN, new CastModifier("byteint"));//$NON-NLS-1$
    	convert.addConvert(FunctionModifier.STRING, FunctionModifier.LONG, new CastModifier("bigint"));//$NON-NLS-1$
    	convert.addConvert(FunctionModifier.STRING, FunctionModifier.SHORT, new CastModifier("smallint"));//$NON-NLS-1$
    	convert.addConvert(FunctionModifier.STRING, FunctionModifier.DOUBLE, new CastModifier("double precision"));//$NON-NLS-1$
    	convert.addConvert(FunctionModifier.STRING, FunctionModifier.BYTE, new CastModifier("byteint")); //$NON-NLS-1$
    	
    	convert.addConvert(FunctionModifier.TIMESTAMP, FunctionModifier.STRING,  new FunctionModifier() {
			@Override
			public List<?> translate(Function function) {
				return Arrays.asList("cast(cast(", function.getParameters().get(0), " AS FORMAT 'Y4-MM-DDBHH:MI:SSDS(6)') AS VARCHAR(26))"); //$NON-NLS-1$ //$NON-NLS-2$
			}
		}); 
    	convert.addConvert(FunctionModifier.TIME, FunctionModifier.STRING,   new FunctionModifier() {
			@Override
			public List<?> translate(Function function) {
				return Arrays.asList("cast(cast(", function.getParameters().get(0), " AS FORMAT 'HH:MI:SS') AS VARCHAR(9))"); //$NON-NLS-1$ //$NON-NLS-2$
			}
		}); 
    	convert.addConvert(FunctionModifier.DATE, FunctionModifier.STRING,  new FunctionModifier() {
			@Override
			public List<?> translate(Function function) {
				return Arrays.asList("cast(cast(", function.getParameters().get(0), " AS FORMAT 'YYYY-MM-DD') AS VARCHAR(11))"); //$NON-NLS-1$ //$NON-NLS-2$
			}
		}); 
    	
    	convert.addTypeMapping("varchar(4000)", FunctionModifier.STRING); //$NON-NLS-1$
    	convert.addNumericBooleanConversions();
		
		registerFunctionModifier(SourceSystemFunctions.CONVERT, convert);
		registerFunctionModifier(SourceSystemFunctions.SUBSTRING, new AliasModifier("substr")); //$NON-NLS-1$	
		registerFunctionModifier(SourceSystemFunctions.LOG, new AliasModifier("LN")); //$NON-NLS-1$
		registerFunctionModifier(SourceSystemFunctions.LOG10, new AliasModifier("LOG"));
		registerFunctionModifier(SourceSystemFunctions.LCASE, new AliasModifier("LOWER")); //$NON-NLS-1$
		registerFunctionModifier(SourceSystemFunctions.UCASE, new AliasModifier("UPPER")); //$NON-NLS-1$
		registerFunctionModifier(SourceSystemFunctions.LENGTH, new AliasModifier("character_length")); //$NON-NLS-1$
		registerFunctionModifier(SourceSystemFunctions.CURDATE, new AliasModifier("CURRENT_DATE")); //$NON-NLS-1$
		registerFunctionModifier(SourceSystemFunctions.CURTIME, new AliasModifier("CURRENT_TIME")); //$NON-NLS-1$
		registerFunctionModifier(SourceSystemFunctions.YEAR, new ExtractModifier("YEAR")); //$NON-NLS-1$
		registerFunctionModifier(SourceSystemFunctions.MONTH, new ExtractModifier("MONTH")); //$NON-NLS-1$
		registerFunctionModifier(SourceSystemFunctions.DAYOFMONTH, new ExtractModifier("DAY")); //$NON-NLS-1$
		registerFunctionModifier(SourceSystemFunctions.HOUR, new ExtractModifier("HOUR")); //$NON-NLS-1$
		registerFunctionModifier(SourceSystemFunctions.MINUTE, new ExtractModifier("MINUTE")); //$NON-NLS-1$
		registerFunctionModifier(SourceSystemFunctions.SECOND, new ExtractModifier("SECOND")); //$NON-NLS-1$
		registerFunctionModifier(SourceSystemFunctions.DAYOFWEEK, new AliasModifier("td_day_of_week"));
		registerFunctionModifier(SourceSystemFunctions.DAYOFYEAR, new AliasModifier("td_day_of_year"));
		registerFunctionModifier(SourceSystemFunctions.QUARTER, new AliasModifier("td_quarter_of_year"));
		registerFunctionModifier(SourceSystemFunctions.WEEK, new AliasModifier("td_week_of_year"));
		registerFunctionModifier(SourceSystemFunctions.CHAR, new AliasModifier("CHR"));
		registerFunctionModifier(SourceSystemFunctions.REPLACE, new AliasModifier("OREPLACE"));
		registerFunctionModifier(SourceSystemFunctions.TRANSLATE, new AliasModifier("OTRANSLATE"));
		registerFunctionModifier(SourceSystemFunctions.IFNULL, new AliasModifier("NVL"));
		registerFunctionModifier(SourceSystemFunctions.CONCAT, new FunctionModifier(){
			@Override
			public List<?> translate(Function function) {
				ArrayList<Object> target = new ArrayList<>();
				target.add("(");
				target.add(function.getParameters().get(0));
				target.add(")||(");
				target.add(function.getParameters().get(1));
				target.add(")");
				return target;
			}
		});
		registerFunctionModifier(SourceSystemFunctions.CONCAT2, new FunctionModifier(){
			@Override
			public List<?> translate(Function function) {
				ArrayList<Object> target = new ArrayList<>();
				target.add("(case when (");
				target.add(function.getParameters().get(0));
				target.add(") is null then ");
				target.add(function.getParameters().get(1));
				target.add("else nvl((");
				target.add(function.getParameters().get(0));
				target.add(")||(");
				target.add(function.getParameters().get(1));
				target.add("), (");
				target.add(function.getParameters().get(0));
				target.add(")) end)");
				return target;
			}
		});
		registerFunctionModifier(SourceSystemFunctions.LOCATE, new LocateModifier(this.convert));
        registerFunctionModifier(SourceSystemFunctions.LEFT, new LeftOrRightFunctionModifier(getLanguageFactory(), this.convert));
        registerFunctionModifier(SourceSystemFunctions.RIGHT, new LeftOrRightFunctionModifier(getLanguageFactory(), this.convert));
        registerFunctionModifier(SourceSystemFunctions.COT, new FunctionModifier() {
			@Override
			public List<?> translate(Function function) {
				function.setName(SourceSystemFunctions.TAN);
				return Collections.singletonList(getLanguageFactory().createFunction(SourceSystemFunctions.DIVIDE_OP, new Expression[]{new Literal(1, TypeFacility.RUNTIME_TYPES.INTEGER), function}, TypeFacility.RUNTIME_TYPES.DOUBLE));
			}
		});        
        registerFunctionModifier(SourceSystemFunctions.LTRIM, new FunctionModifier() {
			@Override
			public List<?> translate(Function function) {
				ArrayList<Object> target = new ArrayList<>();
				target.add("TRIM(LEADING FROM ");//$NON-NLS-1$
				target.add(function.getParameters().get(0));
				target.add(")"); //$NON-NLS-1$				
				return target; 
			}
		}); 
        registerFunctionModifier(SourceSystemFunctions.RTRIM, new FunctionModifier() {
			@Override
			public List<?> translate(Function function) {
				ArrayList<Object> target = new ArrayList<>();
				target.add("TRIM(TRAILING FROM ");//$NON-NLS-1$
				target.add(function.getParameters().get(0));
				target.add(")"); //$NON-NLS-1$
				return target; 
			}
		}); 
        registerFunctionModifier(SourceSystemFunctions.MOD, new FunctionModifier() {
			@Override
			public List<?> translate(Function function) {
				return Arrays.asList(function.getParameters().get(0), " MOD ", function.getParameters().get(1)); //$NON-NLS-1$
			}
		});  
        
        if (supportsRandFunction()) {
	        registerFunctionModifier(SourceSystemFunctions.RAND, new FunctionModifier() {
	 	 	 	@Override
	 	 	 	public List<?> translate(Function function) {
	 	 	 		return Collections.singletonList("(cast(random(0,999999999) as float)/1000000000 (format '9.999999999'))"); //$NON-NLS-1$
	 	 	 	}
	 	 	 });  
        }
        
        addPushDownFunction(TERADATA, "COSH", FLOAT, FLOAT); //$NON-NLS-1$
        addPushDownFunction(TERADATA, "TANH", FLOAT, FLOAT); //$NON-NLS-1$
        addPushDownFunction(TERADATA, "ACOSH", FLOAT, FLOAT); //$NON-NLS-1$
        addPushDownFunction(TERADATA, "ASINH", FLOAT, FLOAT); //$NON-NLS-1$
        addPushDownFunction(TERADATA, "ATANH", FLOAT, FLOAT); //$NON-NLS-1$
        addPushDownFunction(TERADATA, "CHAR2HEXINT", STRING, STRING); //$NON-NLS-1$
        addPushDownFunction(TERADATA, "INDEX", INTEGER, STRING, STRING); //$NON-NLS-1$
        addPushDownFunction(TERADATA, "BYTES", INTEGER, STRING); //$NON-NLS-1$
        addPushDownFunction(TERADATA, "OCTET_LENGTH", INTEGER, STRING); //$NON-NLS-1$
        addPushDownFunction(TERADATA, "HASHAMP", INTEGER, STRING); //$NON-NLS-1$
        addPushDownFunction(TERADATA, "HASHBAKAMP", INTEGER, STRING); //$NON-NLS-1$
        addPushDownFunction(TERADATA, "HASHBUCKET", INTEGER, STRING); //$NON-NLS-1$
        addPushDownFunction(TERADATA, "HASHROW", INTEGER, STRING); //$NON-NLS-1$
        addPushDownFunction(TERADATA, "NULLIFZERO", BIG_DECIMAL, BIG_DECIMAL); //$NON-NLS-1$        
        addPushDownFunction(TERADATA, "ZEROIFNULL", BIG_DECIMAL, BIG_DECIMAL); //$NON-NLS-1$
        registerFunctionModifier(SourceSystemFunctions.REPEAT, new FunctionModifier() {
			@Override
			public List<?> translate(Function function) {
				return Arrays.asList("LPAD(", function.getParameters().get(0), "," , function.getParameters().get(1), "," , function.getParameters().get(0), ")"); //$NON-NLS-1$ //$NON-NLS-2$
			}
		});
		registerFunctionModifier(SourceSystemFunctions.INSERT, new FunctionModifier() {
			@Override
			public List<?> translate(Function function) {
				return Arrays.asList("SUBSTR(", function.getParameters().get(0), "," , 1, "," , function.getParameters().get(1), "-1", ")",
				"||", function.getParameters().get(3), "||", "SUBSTR(", function.getParameters().get(0), ",", function.getParameters().get(1) , "+" , function.getParameters().get(2),
				",CASE WHEN LENGTH(", function.getParameters().get(0), ")-", function.getParameters().get(2), ">=0 THEN LENGTH(", function.getParameters().get(0), ")-", function.getParameters().get(2),
				" ELSE 0 END", ")"); //$NON-NLS-1$ //$NON-NLS-2$
			}
		});
        DVCaller.call(85, this);
	}

	@Override
    public SQLConversionVisitor getSQLConversionVisitor() {
    	return new TeradataSQLConversionVisitor(this);
    }
	
	
    @Override
    public List<String> getSupportedFunctions() {
		List<String> supportedFunctions = new ArrayList<>(super.getSupportedFunctions());

        supportedFunctions.add(SourceSystemFunctions.ABS);
        supportedFunctions.add(SourceSystemFunctions.ACOS);
		supportedFunctions.add(SourceSystemFunctions.ASCII);
        supportedFunctions.add(SourceSystemFunctions.ASIN);
        supportedFunctions.add(SourceSystemFunctions.ATAN);
        supportedFunctions.add(SourceSystemFunctions.ATAN2);
		supportedFunctions.add(SourceSystemFunctions.BITAND);
		supportedFunctions.add(SourceSystemFunctions.BITOR);
		supportedFunctions.add(SourceSystemFunctions.BITXOR);
		supportedFunctions.add(SourceSystemFunctions.BITNOT);
		supportedFunctions.add(SourceSystemFunctions.CEILING);
		supportedFunctions.add(SourceSystemFunctions.CHAR);
        supportedFunctions.add(SourceSystemFunctions.COALESCE);
		supportedFunctions.add(SourceSystemFunctions.CONCAT);
		supportedFunctions.add(SourceSystemFunctions.CONCAT2);
        supportedFunctions.add(SourceSystemFunctions.COS);
        supportedFunctions.add(SourceSystemFunctions.COT);
        supportedFunctions.add(SourceSystemFunctions.CONVERT);
		supportedFunctions.add(SourceSystemFunctions.CURDATE);
		supportedFunctions.add(SourceSystemFunctions.CURTIME); 
		supportedFunctions.add(SourceSystemFunctions.DAYOFMONTH);
		supportedFunctions.add(SourceSystemFunctions.DAYOFWEEK);
		supportedFunctions.add(SourceSystemFunctions.DAYOFYEAR);
		supportedFunctions.add(SourceSystemFunctions.DEGREES);
        supportedFunctions.add(SourceSystemFunctions.EXP);
		supportedFunctions.add(SourceSystemFunctions.FLOOR);
        supportedFunctions.add(SourceSystemFunctions.HOUR);
		supportedFunctions.add(SourceSystemFunctions.IFNULL);
		supportedFunctions.add(SourceSystemFunctions.INITCAP);
		supportedFunctions.add(SourceSystemFunctions.INSERT);
        supportedFunctions.add(SourceSystemFunctions.LEFT);
        supportedFunctions.add(SourceSystemFunctions.LOCATE);
        supportedFunctions.add(SourceSystemFunctions.LOG);
		supportedFunctions.add(SourceSystemFunctions.LOG10);
        supportedFunctions.add(SourceSystemFunctions.LCASE);
		supportedFunctions.add(SourceSystemFunctions.LPAD);
        supportedFunctions.add(SourceSystemFunctions.LTRIM);
        supportedFunctions.add(SourceSystemFunctions.LENGTH);
        supportedFunctions.add(SourceSystemFunctions.MINUTE);
        supportedFunctions.add(SourceSystemFunctions.MOD);
        supportedFunctions.add(SourceSystemFunctions.MONTH);
        supportedFunctions.add(SourceSystemFunctions.NULLIF);
		supportedFunctions.add(SourceSystemFunctions.POWER);
		supportedFunctions.add(SourceSystemFunctions.QUARTER);
		supportedFunctions.add(SourceSystemFunctions.RADIANS);
		supportedFunctions.add(SourceSystemFunctions.REPEAT);
		supportedFunctions.add(SourceSystemFunctions.REPLACE);
        supportedFunctions.add(SourceSystemFunctions.RIGHT);
        supportedFunctions.add(SourceSystemFunctions.ROUND);
		supportedFunctions.add(SourceSystemFunctions.RPAD);
        supportedFunctions.add(SourceSystemFunctions.RTRIM);
		supportedFunctions.add(SourceSystemFunctions.SECOND);
		supportedFunctions.add(SourceSystemFunctions.SIGN);
        supportedFunctions.add(SourceSystemFunctions.SIN);
        supportedFunctions.add(SourceSystemFunctions.SQRT);
        supportedFunctions.add(SourceSystemFunctions.SUBSTRING);
        supportedFunctions.add(SourceSystemFunctions.TAN);
		supportedFunctions.add(SourceSystemFunctions.TRANSLATE);
        supportedFunctions.add(SourceSystemFunctions.TRIM);
        supportedFunctions.add(SourceSystemFunctions.UCASE);
		supportedFunctions.add(SourceSystemFunctions.WEEK);
        supportedFunctions.add(SourceSystemFunctions.YEAR);
        if (supportsRandFunction()) {
        	supportedFunctions.add(SourceSystemFunctions.RAND);
        }

        return supportedFunctions;
    }
    
    @Override
    public String translateLiteralDate(Date dateValue) {
        return "cast('" + formatDateValue(dateValue) + "' AS DATE FORMAT 'yyyy-mm-dd')"; //$NON-NLS-1$//$NON-NLS-2$
    }

    @Override
    public String translateLiteralTime(Time timeValue) {
        return "cast('" + formatDateValue(timeValue) + "' AS TIME(0) FORMAT 'hh:mi:ss')"; //$NON-NLS-1$//$NON-NLS-2$
    }
    
    @Override
    public String translateLiteralTimestamp(Timestamp timestampValue) {
        return "cast('" + formatDateValue(timestampValue) + "' AS TIMESTAMP(6))"; //$NON-NLS-1$//$NON-NLS-2$ 
    }	
    
    // Teradata also supports MINUS & ALL set operators
    // more aggregates available

    @Override
    public boolean supportsScalarSubqueries() {
        return false;
    }
    
    @Override
    public boolean supportsUnions() {
    	return true;
    }

    @Override
    public boolean supportsIntersect() {
    	return true;
    }

    @Override
    public boolean supportsExcept() {
    	return true;
    }    
    
    @Override
    public boolean supportsInlineViews() {
        return true;
    }  
    
    @Override
    public boolean supportsAggregatesEnhancedNumeric() {
    	return true;
    }
    
    @Override
    public boolean supportsCommonTableExpressions() {
    	return false;
    }

	@Override
	public boolean supportsElementaryOlapOperations() {
		return true;
	}    
    
    @Override
    public List<?> translateLimit(Limit limit, ExecutionContext context) {
    	return Arrays.asList("TOP ", limit.getRowLimit()); //$NON-NLS-1$
    }
    
    @Override
    public boolean useSelectLimit() {
    	return true;
    }
    
    @Override
    public boolean supportsRowLimit() {
        return true;
    }
    
    @Override
    public NullOrder getDefaultNullOrder() {
    	return NullOrder.FIRST;
    }
    
    @Override
    public List<?> translateCommand(Command command, ExecutionContext context) {
    	if (command instanceof QueryExpression) {
    		QueryExpression qe = (QueryExpression)command;
    		//teradata prefers positional ordering
    		if (qe.getOrderBy() != null) {
    			Select select = qe.getProjectedQuery();
    			List<DerivedColumn> derivedColumns = select.getDerivedColumns();
    			Map<String, Integer> positions = new HashMap<>();
    			int i = 1;
    			for (DerivedColumn derivedColumn : derivedColumns) {
    				String name = derivedColumn.getAlias();
    				if (name == null && derivedColumn.getExpression() instanceof ColumnReference) {
    					ColumnReference cr = (ColumnReference)derivedColumn.getExpression();
    					name = cr.toString();
    				}
					positions.put(name, i++);
				}
    			for (SortSpecification ss : qe.getOrderBy().getSortSpecifications()) {
    				Expression ex = ss.getExpression();
    				if (!(ex instanceof ColumnReference)) {
    					continue;
    				} 
    				ColumnReference cr = (ColumnReference)ex;
    				Integer position = positions.get(cr.toString());
    				if (position != null) {
    					ss.setExpression(new Literal(position, TypeFacility.RUNTIME_TYPES.INTEGER));
    				}
				}
    		}
    	}
    	return super.translateCommand(command, context);
    }
    
    public static class LocateModifier extends FunctionModifier {
    	ConvertModifier convertModifier;
    	
    	public LocateModifier(ConvertModifier convertModifier) {
    		this.convertModifier = convertModifier;
    	}
		
    	@Override
		public List<?> translate(Function function) {
    		ArrayList<Object> target = new ArrayList<>();
    		Expression expr1 =  function.getParameters().get(0);
    		Expression expr2 =  function.getParameters().get(1);
    		if (function.getParameters().size() > 2) {
    			Expression expr3 =  function.getParameters().get(2);
	    		target.add("position("); //$NON-NLS-1$
	    		target.add(expr1);
	    		target.add( " in "); //$NON-NLS-1$
	    		target.add("substr("); //$NON-NLS-1$
	    		target.add(expr2);
	    		target.add(","); //$NON-NLS-1$
	    		target.add(expr3);
	    		target.add("))"); //$NON-NLS-1$	    		
    		}
    		else {
	    		target.add("position("); //$NON-NLS-1$
	    		target.add(expr1);
	    		target.add( " in "); //$NON-NLS-1$
	    		target.add(expr2);
	    		target.add(")"); //$NON-NLS-1$
    		}
    		return target;
		}
	}
    
	public static class ExtractModifier extends FunctionModifier {
    	private String type;
    	public ExtractModifier(String type) {
    		this.type = type;
    	}
		@Override
		public List<?> translate(Function function) {
			return Arrays.asList("extract(",this.type," from ",function.getParameters().get(0) ,")"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ 				
		}
	}
    
    public static class CastModifier extends FunctionModifier {
    	private String target;
    	public CastModifier(String target) {
    		this.target = target;
    	}
		@Override
		public List<?> translate(Function function) {
			return Arrays.asList("cast(", function.getParameters().get(0), " AS "+this.target+")"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		}
	}
    
    public static class LeftOrRightFunctionModifier extends FunctionModifier {
        private LanguageFactory langFactory;
        ConvertModifier convertModifier;
        
        public LeftOrRightFunctionModifier(LanguageFactory langFactory, ConvertModifier converModifier) {
            this.langFactory = langFactory;
            this.convertModifier = converModifier;
        }
        
        @Override
        public List<?> translate(Function function) {
            List<Expression> args = function.getParameters();
            ArrayList<Object> target = new ArrayList<>();
            if (function.getName().equalsIgnoreCase("left")) { //$NON-NLS-1$
            	//substr(string, 1, length)
            	target.add("substr("); //$NON-NLS-1$
            	target.add(args.get(0));
            	target.add(","); //$NON-NLS-1$
            	target.add(langFactory.createLiteral(1, TypeFacility.RUNTIME_TYPES.INTEGER));
            	target.add(","); //$NON-NLS-1$
            	target.add(args.get(1));
            	target.add(")"); //$NON-NLS-1$
            } else if (function.getName().equalsIgnoreCase("right")) { //$NON-NLS-1$
            	//substr(case_size, character_length(case_size) -4) 
            	target.add("substr("); //$NON-NLS-1$
            	target.add(args.get(0));
            	
            	target.add(",(character_length("); //$NON-NLS-1$
            	target.add(args.get(0));
            	target.add(")-"); //$NON-NLS-1$
            	target.add(args.get(1));
            	target.add("+1))"); //$NON-NLS-1$ // offset for 1 based index
            }
            return target;
        }
    }    

	public ITypeConverter getTypeConverter() {
		return new ITypeConverter() {
			@Override
			public String getRuntimeType(int type, String typeName, int precision, int scale) {
				if ("JSON".equals(typeName)) {
					return DataTypeManager.DefaultDataTypes.CLOB;
				}
				if ("CHAR".equals(typeName) || "NCHAR".equals(typeName)) {
					if (precision == 1)
						return DataTypeManager.DefaultDataTypes.CHAR;
					return DataTypeManager.DefaultDataTypes.STRING;
				}
				if (typeOfTypeEq(typeName, "PERIOD", null))
					return DataTypeManager.DefaultDataTypes.STRING;
				if (type == Types.FLOAT || type == Types.DOUBLE) {
					return DataTypeManager.DefaultDataTypes.DOUBLE;
				}
				if (typeName != null && typeName.toUpperCase().startsWith("INTERVAL"))
					return DataTypeManager.DefaultDataTypes.STRING;
				return null;
			}

			@Override
			public boolean filterColumn(String columnName, int type, String typeName, int precision, int scale, int charOctetLength) {
				return type != 1111 || !(typeOfTypeEq(typeName, "PERIOD", null) && charOctetLength == 0);
			}

			@Override
			protected String usualNumericCheck(String typeName, int precision, int scale) {
				String type = noBooleanNumericCheck(typeName, precision, scale);
				if (type != null)
					return type;
				return super.usualNumericCheck(typeName, precision, scale);
			}

			@Override
			public void tuneColumn(Column column, int type, String typeName, int precision, int scale) {
				if (typeOfTypeEq(typeName, "PERIOD", null)) {
					column.setLength(255);
					column.setPrecision(255);
				}
				if ("NUMBER".equals(typeName) && precision > 38) {
					if (scale == 0) {
						column.setRuntimeType(DataTypeManager.DefaultDataTypes.BIG_DECIMAL);
						column.setPrecision(NUMBER_PREC);
						column.setScale(NUMBER_SCALE);
						return;
					} else
						column.setPrecision(38);
				}
				super.tuneColumn(column, type, typeName, precision, scale);
			}
		};
	}
	
	@Override
    protected boolean supportsBooleanExpressions() {
        return false;
    }

	public String translateLiteralBinaryType(BinaryType obj) {
		return "'" + obj + "'XBV";
	}

	@Override
	public void bindValue(PreparedStatement stmt, Object param, Class<?> paramType, int i) throws SQLException {
		if (param instanceof Clob) {
			Clob clob = (Clob)param;
			stmt.setCharacterStream(i, clob.getCharacterStream(), clob.length());
			return;
		}
		if (param instanceof Blob) {
			Blob blob = (Blob)param;
			stmt.setBinaryStream(i, blob.getBinaryStream(), blob.length());
			return;
		}
		if (param instanceof SQLXML) {
			SQLXML src = (SQLXML)param;
			SQLXML trg = stmt.getConnection().createSQLXML();
			try {
				IOUtils.copy(src.getCharacterStream(), trg.setCharacterStream());
			} catch (IOException e) {
				throw new SQLException(e);
			}
			stmt.setSQLXML(i, trg);
			return;
		}
		if (Boolean.class.isAssignableFrom(paramType)) {
			if (param == null)
				stmt.setNull(i, Types.INTEGER);
			else
				stmt.setInt(i, ((Boolean)param) ? 1 : 0);
			return;
		}
		if (param == null) {
			if (Time.class.isAssignableFrom(paramType)) {
				stmt.setTime(i, null, getDatabaseCalendar());
				return;
			}
			if (Timestamp.class.isAssignableFrom(paramType)) {
				stmt.setTimestamp(i, null, getDatabaseCalendar());
				return;
			}
		}
		if (paramType.equals(TypeFacility.RUNTIME_TYPES.DATE)) {
			stmt.setDate(i,(java.sql.Date)param);
			return;
		}
		if (paramType.equals(TypeFacility.RUNTIME_TYPES.TIME)) {
			stmt.setTime(i,(java.sql.Time)param);
			return;
		}
		if (paramType.equals(TypeFacility.RUNTIME_TYPES.TIMESTAMP)) {
			stmt.setTimestamp(i,(java.sql.Timestamp)param);
			return;
		}
		super.bindValue(stmt, param, paramType, i);
	}

	@Override
	public Object retrieveValue(ResultSet results, int columnIndex, Class<?> expectedType) throws SQLException {
		Object val = super.retrieveValue(results, columnIndex, expectedType);
		if (val instanceof Struct)
			val = val.toString();
		if (TypeFacility.RUNTIME_TYPES.TIME.equals(expectedType)) {
			return results.getTime(columnIndex);
		}
		if (TypeFacility.RUNTIME_TYPES.DATE.equals(expectedType)) {
			return results.getDate(columnIndex);
		}
		if (TypeFacility.RUNTIME_TYPES.TIMESTAMP.equals(expectedType)) {
			return results.getTimestamp(columnIndex);
		}
		return val;
	}

	@Override
	public Object retrieveValue(CallableStatement results, int parameterIndex, Class<?> expectedType) throws SQLException {
		Object val = super.retrieveValue(results, parameterIndex, expectedType);
		if (val instanceof Struct)
			val = val.toString();
		if (TypeFacility.RUNTIME_TYPES.TIME.equals(expectedType)) {
			return results.getTime(parameterIndex);
		}
		if (TypeFacility.RUNTIME_TYPES.DATE.equals(expectedType)) {
			return results.getDate(parameterIndex);
		}
		if (TypeFacility.RUNTIME_TYPES.TIMESTAMP.equals(expectedType)) {
			return results.getTimestamp(parameterIndex);
		}
		return val;
	}
}
