/*
 * JBoss, Home of Professional Open Source.
 * See the COPYRIGHT.txt file distributed with this work for information
 * regarding copyright ownership.  Some portions may be licensed
 * to Red Hat, Inc. under one or more contributor license agreements.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA.
 * Modified by
 * @author  Data Virtuality GmbH
 * @date    $Date$
 *  
 */
 
// https://issues.jboss.org/browse/TEIID-2476
// https://issues.jboss.org/browse/TEIID-5516
package org.teiid.dqp.internal.process;

import com.datavirtuality.dv.core.interfaces.DVCaller;
import org.teiid.PolicyDecider;
import org.teiid.adminapi.DataPolicy.Context;
import org.teiid.adminapi.DataPolicy.PermissionType;
import org.teiid.api.exception.query.QueryValidatorException;
import org.teiid.core.CoreConstants;
import org.teiid.core.TeiidComponentException;
import org.teiid.metadata.AbstractMetadataRecord;
import org.teiid.metadata.Procedure;
import org.teiid.metadata.Schema;
import org.teiid.query.metadata.QueryMetadataInterface;
import org.teiid.query.sql.lang.Command;
import org.teiid.query.util.CommandContext;

import java.util.Collection;
import java.util.HashSet;

/**
 * The default Teiid authorization validator
 */
public class DefaultAuthorizationValidator implements AuthorizationValidator {
	
	private PolicyDecider policyDecider;
	
	public DefaultAuthorizationValidator() {
	}

	@Override
	public boolean validate(String[] originalSql, Command command,
			QueryMetadataInterface metadata, CommandContext commandContext,
			CommandType commandType) throws QueryValidatorException,
			TeiidComponentException {
		if (policyDecider != null && policyDecider.validateCommand(commandContext)) {
			validateWithVisitor(command, metadata, commandContext);
		}
		return false;
	}

	private void validateWithVisitor(Command command, QueryMetadataInterface metadata, CommandContext commandContext)
			throws QueryValidatorException, TeiidComponentException {
		// EXTENSION
		// ISSUE: DVCORE-716
		// FIXED BY: Roman Pedchenko
		// DATE: 2013.10.24
//			AuthorizationValidationVisitor visitor = new AuthorizationValidationVisitor(this.policyDecider, commandContext);
		AuthorizationValidationVisitor visitor = (AuthorizationValidationVisitor) DVCaller.call(28, this.policyDecider, commandContext);
		if (visitor == null)
			visitor = new AuthorizationValidationVisitor(this.policyDecider, commandContext);
		// END EXTENSION
		Request.validateWithVisitor(visitor, metadata, command);
	}
	
	@Override
	public boolean hasRole(String roleName, CommandContext commandContext) {
		if (policyDecider == null) {
			return true;
		}
		return this.policyDecider.hasRole(roleName, commandContext);
	}
	
	public void setPolicyDecider(PolicyDecider policyDecider) {
		this.policyDecider = policyDecider;
	}
	
	public PolicyDecider getPolicyDecider() {
		return policyDecider;
	}
	
	@Override
 	public boolean isAccessible(AbstractMetadataRecord record,
 		CommandContext commandContext) {
		if (policyDecider == null || !policyDecider.validateCommand(commandContext)) {
			return true;
		}
		AbstractMetadataRecord parent = record;
		while (parent.getParent() != null) {
			parent = parent.getParent();
			if (parent instanceof Procedure) {
				return true; //don't check procedure params/rs columns
			}
		}
		if (!(parent instanceof Schema) || (CoreConstants.SYSTEM_MODEL.equalsIgnoreCase(parent.getName()) || CoreConstants.ODBC_MODEL.equalsIgnoreCase(parent.getName()))) {
			//access is always allowed to system tables / procedures or unrooted objects
			return true;
		}
		PermissionType action = PermissionType.READ;
//		DVCORE-8142
//		if (record instanceof FunctionMethod || record instanceof Procedure) {
//			action = PermissionType.EXECUTE;
//		}

		//cache permission check
		Boolean result = isAccessible(record, commandContext, action);
		if (!result && record instanceof Schema) {
			Schema s = (Schema)record;
			//check if anything can be seen below the schema level
	 	
			if (anyAccessible(s.getTables().values(), commandContext, PermissionType.READ)) {
				result = true;
//			DVCORE-8142
//			} else if (anyAccessible(s.getProcedures().values(), commandContext, PermissionType.EXECUTE)) {
			} else if (anyAccessible(s.getProcedures().values(), commandContext, PermissionType.READ)) {
				result = true;
//			DVCORE-8142
//			} else if (anyAccessible(s.getFunctions().values(), commandContext, PermissionType.EXECUTE)) {
			} else if (anyAccessible(s.getFunctions().values(), commandContext, PermissionType.READ)) {
				result = true;
			}
	 	}
	 	commandContext.setAccessible(record, result);
	 	return result;
	 }

	private boolean anyAccessible(Collection<? extends AbstractMetadataRecord> records, CommandContext commandContext, PermissionType permission) {
		for (AbstractMetadataRecord record : records) {
	 		if (isAccessible(record, commandContext, permission)) {
	 			return true;
	 		}
	 	}
	 	return false;
	}

	private Boolean isAccessible(AbstractMetadataRecord record,
			CommandContext commandContext, PermissionType action) {
		Boolean result = commandContext.isAccessible(record);
		if (result != null) {
			return result;
		}

		HashSet<String> resources = new HashSet<String>(2);
		resources.add(record.getFullName());
		return this.policyDecider.getInaccessibleResources(action, resources, Context.METADATA, commandContext).isEmpty();
	}


	@Override
	public boolean isAccessible(String fullName,
								CommandContext commandContext, PermissionType action) {
		HashSet<String> resources = new HashSet<>(2);
		resources.add(fullName);
		return this.policyDecider.getInaccessibleResources(action, resources, Context.METADATA, commandContext).isEmpty();
	}
}
