/*
 * JBoss, Home of Professional Open Source.
 * See the COPYRIGHT.txt file distributed with this work for information
 * regarding copyright ownership.  Some portions may be licensed
 * to Red Hat, Inc. under one or more contributor license agreements.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA.
 * Modified by
 * @author  Data Virtuality GmbH
 * @date    $Date$
 *  
 */
/*
 EXTENSION
 ISSUE: DVCORE-2606
 FIXED BY: Salvatore Raunich
 DATE: 2015-02-11
 ADOPTED FROM: TEIID-3328
 NOTES: DVCORE-2606 has been fixed by partially backporting TEIID-3328, since the patch provided in TEIID-3328 was based
 		on source code introduced in other issues and never backported in our current code.
 END EXTENSION
 */
// https://issues.jboss.org/browse/TEIID-3716
// https://issues.jboss.org/browse/TEIID-3846
// https://issues.jboss.org/browse/TEIID-3101
// https://issues.jboss.org/browse/TEIID-2085
// https://issues.jboss.org/browse/TEIID-2311
// https://issues.jboss.org/browse/TEIID-5146
// https://issues.jboss.org/browse/TEIID-5152
// https://issues.jboss.org/browse/TEIID-4076
// https://issues.jboss.org/browse/TEIID-2245
// https://issues.jboss.org/browse/TEIID-5235
// https://issues.jboss.org/browse/TEIID-3325
// https://issues.jboss.org/browse/TEIID-2309
// https://issues.jboss.org/browse/TEIID-2381
// https://issues.jboss.org/browse/TEIID-2555
// https://issues.jboss.org/browse/TEIID-3530
// https://issues.jboss.org/browse/TEIID-3630
// https://issues.redhat.com/browse/TEIID-2591
// https://issues.redhat.com/browse/TEIID-3119
package org.teiid.query.optimizer.relational;

import java.util.*;
import java.util.concurrent.atomic.AtomicInteger;

import org.teiid.adminapi.impl.ModelMetaData;
import org.teiid.adminapi.impl.VDBMetaData;
import org.teiid.api.exception.query.QueryMetadataException;
import org.teiid.api.exception.query.QueryPlannerException;
import org.teiid.core.CoreConstants;
import org.teiid.core.TeiidComponentException;
import org.teiid.core.TeiidRuntimeException;
import org.teiid.core.id.IDGenerator;
import org.teiid.core.util.Assertion;
import org.teiid.core.util.StringUtil;
import org.teiid.metadata.AbstractMetadataRecord;
import org.teiid.metadata.FunctionMethod.Determinism;
import org.teiid.metadata.Table;
import org.teiid.query.QueryPlugin;
import org.teiid.query.analysis.AnalysisRecord;
import org.teiid.query.metadata.QueryMetadataInterface;
import org.teiid.query.metadata.TempMetadataAdapter;
import org.teiid.query.metadata.TempMetadataID;
import org.teiid.query.optimizer.capabilities.CapabilitiesFinder;
import org.teiid.query.optimizer.capabilities.SourceCapabilities;
import org.teiid.query.optimizer.capabilities.SourceCapabilities.Capability;
import org.teiid.query.optimizer.relational.plantree.NodeConstants;
import org.teiid.query.optimizer.relational.plantree.NodeConstants.Info;
import org.teiid.query.optimizer.relational.plantree.PlanNode;
import org.teiid.query.optimizer.relational.rules.CapabilitiesUtil;
import org.teiid.query.optimizer.relational.rules.NewCalculateCostUtil;
import org.teiid.query.optimizer.relational.rules.RuleAssignOutputElements;
import org.teiid.query.optimizer.relational.rules.RuleChooseJoinStrategy;
import org.teiid.query.processor.ProcessorPlan;
import org.teiid.query.processor.RegisterRequestParameter;
import org.teiid.query.processor.relational.*;
import org.teiid.query.processor.relational.JoinNode.JoinStrategyType;
import org.teiid.query.processor.relational.MergeJoinStrategy.SortOption;
import org.teiid.query.processor.relational.SortUtility.Mode;
import org.teiid.query.resolver.util.ResolverUtil;
import org.teiid.query.sql.lang.*;
import org.teiid.query.sql.lang.ObjectTable.ObjectColumn;
import org.teiid.query.sql.lang.SetQuery.Operation;
import org.teiid.query.sql.lang.SourceHint.SpecificHint;
import org.teiid.query.sql.lang.XMLTable.XMLColumn;
import org.teiid.query.sql.symbol.ElementSymbol;
import org.teiid.query.sql.symbol.Expression;
import org.teiid.query.sql.symbol.ExpressionSymbol;
import org.teiid.query.sql.symbol.GroupSymbol;
import org.teiid.query.sql.symbol.Reference;
import org.teiid.query.sql.symbol.WindowFunction;
import org.teiid.query.sql.util.SymbolMap;
import org.teiid.query.sql.visitor.ElementCollectorVisitor;
import org.teiid.query.sql.visitor.EvaluatableVisitor;
import org.teiid.query.sql.visitor.ReferenceCollectorVisitor;
import org.teiid.query.sql.visitor.EvaluatableVisitor.EvaluationLevel;
import org.teiid.query.sql.visitor.GroupCollectorVisitor;
import org.teiid.query.util.CommandContext;
import org.teiid.translator.ExecutionFactory.TransactionSupport;

import com.datavirtuality.dv.core.interfaces.DVCaller;
import com.datavirtuality.dv.core.interfaces.query.sql.lang.SelectInto;


public class PlanToProcessConverter {
	protected QueryMetadataInterface metadata;
	private IDGenerator idGenerator;
	private AnalysisRecord analysisRecord;
	private CapabilitiesFinder capFinder;
	
	//state for detecting and reusing source queries
	private Map<Command, AccessNode> sharedCommands = new HashMap<Command, AccessNode>();
	private CommandContext context;
	private static AtomicInteger sharedId = new AtomicInteger();
	
	public static class SharedStateKey {
		int id;
		int expectedReaders;
	}

	public PlanToProcessConverter(QueryMetadataInterface metadata, IDGenerator idGenerator, AnalysisRecord analysisRecord, CapabilitiesFinder capFinder, CommandContext context) {
		this.metadata = metadata;
		this.idGenerator = idGenerator;
		this.analysisRecord = analysisRecord;
		this.capFinder = capFinder;
		this.context = context;
	}
	
    public RelationalPlan convert(PlanNode planNode)
        throws QueryPlannerException, TeiidComponentException {
    	try {
	        boolean debug = analysisRecord.recordDebug();
	        if(debug) {
	            analysisRecord.println("\n============================================================================"); //$NON-NLS-1$
	            analysisRecord.println("CONVERTING PLAN TREE TO PROCESS TREE"); //$NON-NLS-1$
	        }
	
	        // Convert plan tree nodes into process tree nodes
	        RelationalNode processNode = convertPlan(planNode);
	        if(debug) {
	            analysisRecord.println("\nPROCESS PLAN = \n" + processNode); //$NON-NLS-1$
	            analysisRecord.println("============================================================================"); //$NON-NLS-1$
	        }
	
	        RelationalPlan processPlan = new RelationalPlan(processNode);
	        return processPlan;
    	} finally {
    		sharedCommands.clear();
    	}
    }

	private RelationalNode convertPlan(PlanNode planNode)
		throws QueryPlannerException, TeiidComponentException {

		// Convert current node in planTree
		RelationalNode convertedNode = convertNode(planNode);
		
		if(convertedNode == null) {
		    Assertion.assertTrue(planNode.getChildCount() == 1);
	        return convertPlan(planNode.getFirstChild());
		}
		
		RelationalNode nextParent = convertedNode;
		
        // convertedNode may be the head of 1 or more nodes   - go to end of chain
        while(nextParent.getChildren()[0] != null) {
            nextParent = nextParent.getChildren()[0];
        }
		
		// Call convertPlan recursively on children
		for (PlanNode childNode : planNode.getChildren()) {
			RelationalNode child = convertPlan(childNode);
			if (planNode.getType() == NodeConstants.Types.SET_OP && nextParent instanceof UnionAllNode
						&& childNode.getProperty(Info.SET_OPERATION) == childNode.getProperty(Info.SET_OPERATION)
						&& childNode.getType() == NodeConstants.Types.SET_OP && childNode.hasBooleanProperty(Info.USE_ALL)) {
				for (RelationalNode grandChild : child.getChildren()) {
					if (grandChild != null) {
						nextParent.addChild(grandChild);
					}
				}
			} else {
				nextParent.addChild(child);
			}
		}

        // Return root of tree for top node
		return convertedNode;
	}

	protected int getID() {
		return idGenerator.nextInt();
	}

	private boolean checkStreamUpdate(Object modelID) throws TeiidComponentException {
		String modelName = metadata.getFullName(modelID);
		try {
			SourceCapabilities caps = capFinder.findCapabilities(modelName);
			if (!caps.supportsCapability(Capability.STREAM_UPDATE)) {
				return false;
			}
		} catch (TeiidRuntimeException e) {
			return false;
		}
		return true;
	}

	private boolean checkStreamPossibleUnary(UnaryFromClause from) throws TeiidComponentException {
		Object groupID = from.getGroup().getMetadataID();
		Object modelID = metadata.getModelID(groupID);
		return checkStreamUpdate(modelID);
	}

	private boolean checkStreamPossibleSubquery(SubqueryFromClause clause) throws TeiidComponentException {
		final Command command = clause.getCommand();
		if (command instanceof StoredProcedure) {
			StoredProcedure call = (StoredProcedure) command;
			return checkStreamUpdate(call.getModelID());
		}
		return false;
	}

	private boolean checkStreamPossibleFrom(FromClause clause) throws TeiidComponentException {
		if (clause instanceof ArrayTable)
			return true;
		if (clause instanceof UnaryFromClause && checkStreamPossibleUnary((UnaryFromClause) clause))
			return true;
		if (clause instanceof SubqueryFromClause && checkStreamPossibleSubquery((SubqueryFromClause) clause))
			return true;
		if (clause instanceof JoinPredicate) {
			JoinPredicate join = (JoinPredicate) clause;
			if (!checkStreamPossibleFrom(join.getLeftClause()))
				return false;
			return checkStreamPossibleFrom(join.getRightClause());
		}
		return false;
	}

	protected RelationalNode convertNode(PlanNode node)
		throws QueryPlannerException, TeiidComponentException {

		RelationalNode processNode = null;

		switch(node.getType()) {
			case NodeConstants.Types.PROJECT:
                GroupSymbol intoGroup = (GroupSymbol) node.getProperty(NodeConstants.Info.INTO_GROUP);
                if(intoGroup != null) {
                    try {
                    	Insert insert = (Insert)node.getFirstChild().getProperty(Info.VIRTUAL_COMMAND);
                        List<ElementSymbol> allIntoElements = insert.getVariables();
                        
                        Object groupID = intoGroup.getMetadataID();
                        Object modelID = metadata.getModelID(groupID);
                        String modelName = metadata.getFullName(modelID);
                        if (metadata.isVirtualGroup(groupID)) {
                        	InsertPlanExecutionNode ipen = new InsertPlanExecutionNode(getID(), metadata);
                        	ipen.setProcessorPlan((ProcessorPlan)node.getFirstChild().getProperty(Info.PROCESSOR_PLAN));
                        	ipen.setReferences(insert.getValues());
                        	processNode = ipen;
                        } else {
	                        ProjectIntoNode pinode = new ProjectIntoNode(getID());
	                        pinode.setIntoGroup(intoGroup);
	                        pinode.setIntoElements(allIntoElements);
	                        pinode.setModelName(modelName);
	                        pinode.setConstraint((Criteria) node.getProperty(Info.CONSTRAINT));
	                        pinode.setVirtualCommand(insert);
	                        if (node.hasBooleanProperty(Info.UPSERT)) {
		                        pinode.setUpsert(true);
	                        }
	                        processNode = pinode;
                            SourceCapabilities caps = capFinder.findCapabilities(modelName);
                            boolean streamPossible = false;
                            if (insert.getQueryExpression() != null && insert.getQueryExpression() instanceof Query
                            	&& ((Query) insert.getQueryExpression()).getFrom() != null) {
                            		streamPossible = true;
                            		for (FromClause clause : ((Query) insert.getQueryExpression()).getFrom().getClauses()) {
                            			if (checkStreamPossibleFrom(clause))
                            				continue;
										streamPossible = false;
										break;
                            		}
                            	}
                            if (streamPossible && caps.supportsCapability(Capability.STREAM_UPDATE)) {
                            	pinode.setMode(org.teiid.query.processor.relational.ProjectIntoNode.Mode.STREAM);
                            } else if (caps.supportsCapability(Capability.INSERT_WITH_ITERATOR)) {
                                	pinode.setMode(org.teiid.query.processor.relational.ProjectIntoNode.Mode.ITERATOR);
                            } else if (caps.supportsCapability(Capability.BATCHED_UPDATES)) {
                            	pinode.setMode(org.teiid.query.processor.relational.ProjectIntoNode.Mode.BATCH);
                            } else {
                            	pinode.setMode(org.teiid.query.processor.relational.ProjectIntoNode.Mode.SINGLE);
                            }
                            pinode.setTransactionSupport((TransactionSupport) caps.getSourceProperty(Capability.TRANSACTION_SUPPORT));
                            if (caps.getSourceProperty(Capability.OEVRRIDE_FETCH_SIZE) != null)
                            	pinode.setOverrideFetchSize((Integer) caps.getSourceProperty(Capability.OEVRRIDE_FETCH_SIZE));
                        }
                    } catch(QueryMetadataException e) {
                         throw new TeiidComponentException(QueryPlugin.Event.TEIID30247, e);
                    }

                } else {
                    List<Expression> symbols = (List) node.getProperty(NodeConstants.Info.PROJECT_COLS);
                    
                    ProjectNode pnode = new ProjectNode(getID());
                    pnode.setSelectSymbols(symbols);
            		processNode = pnode;
            		
               		if (node.hasBooleanProperty(Info.HAS_WINDOW_FUNCTIONS)) {
            			WindowFunctionProjectNode wfpn = new WindowFunctionProjectNode(getID());
            			Set<WindowFunction> windowFunctions = RuleAssignOutputElements.getWindowFunctions(symbols);
            			//TODO: check for selecting all window functions
            			List<Expression> outputElements = new ArrayList<Expression>(windowFunctions);
            			//collect the other projected expressions
            			for (Expression singleElementSymbol : (List<Expression>)node.getFirstChild().getProperty(Info.OUTPUT_COLS)) {
							outputElements.add(singleElementSymbol);
						}
            			wfpn.setElements(outputElements);
            			wfpn.init();
            			pnode.addChild(wfpn);
            		}
                }
                break;

			case NodeConstants.Types.JOIN:
                JoinType jtype = (JoinType) node.getProperty(NodeConstants.Info.JOIN_TYPE);
                JoinStrategyType stype = (JoinStrategyType) node.getProperty(NodeConstants.Info.JOIN_STRATEGY);

                JoinNode jnode = new JoinNode(getID());
                jnode.setJoinType(jtype);
                jnode.setLeftDistinct(node.hasBooleanProperty(NodeConstants.Info.IS_LEFT_DISTINCT));
                jnode.setRightDistinct(node.hasBooleanProperty(NodeConstants.Info.IS_RIGHT_DISTINCT));
                List joinCrits = (List) node.getProperty(NodeConstants.Info.JOIN_CRITERIA);
                String depValueSource = (String) node.getProperty(NodeConstants.Info.DEPENDENT_VALUE_SOURCE);
                boolean dvDep = node.hasBooleanProperty(NodeConstants.Info.DV_DEP);
                SortOption leftSort = (SortOption)node.getProperty(NodeConstants.Info.SORT_LEFT);
                if(stype == JoinStrategyType.MERGE || stype == JoinStrategyType.ENHANCED_SORT) {
                	MergeJoinStrategy mjStrategy = null;
                	if (stype.equals(JoinStrategyType.ENHANCED_SORT)) { 
                		EnhancedSortMergeJoinStrategy esmjStrategy = new EnhancedSortMergeJoinStrategy(leftSort, (SortOption)node.getProperty(NodeConstants.Info.SORT_RIGHT));
                		esmjStrategy.setSemiDep(node.hasBooleanProperty(Info.IS_SEMI_DEP));
                		mjStrategy = esmjStrategy;
                	} else {
                		mjStrategy = new MergeJoinStrategy(leftSort, (SortOption)node.getProperty(NodeConstants.Info.SORT_RIGHT), false);
                	}
                    jnode.setJoinStrategy(mjStrategy);
                    List leftExpressions = (List) node.getProperty(NodeConstants.Info.LEFT_EXPRESSIONS);
                    List rightExpressions = (List) node.getProperty(NodeConstants.Info.RIGHT_EXPRESSIONS);
                    jnode.setJoinExpressions(leftExpressions, rightExpressions);
                    joinCrits = (List) node.getProperty(NodeConstants.Info.NON_EQUI_JOIN_CRITERIA);
                } else if (stype == JoinStrategyType.NESTED_TABLE || dvDep) {
                	NestedTableJoinStrategy ntjStrategy = new NestedTableJoinStrategy();
                	jnode.setJoinStrategy(ntjStrategy);
	                SymbolMap references = (SymbolMap)node.getProperty(Info.RIGHT_NESTED_REFERENCES);
                	ntjStrategy.setRightMap(references);
                } else {
                    NestedLoopJoinStrategy nljStrategy = new NestedLoopJoinStrategy();
                    jnode.setJoinStrategy(nljStrategy);
                }
                Criteria joinCrit = Criteria.combineCriteria(joinCrits);
                jnode.setJoinCriteria(joinCrit);
                               
                processNode = jnode;
                
                jnode.setDependentValueSource(depValueSource);
                
				break;

			case NodeConstants.Types.ACCESS:
                ProcessorPlan plan = (ProcessorPlan) node.getProperty(NodeConstants.Info.PROCESSOR_PLAN);
                if(plan != null) {
                    
                    PlanExecutionNode peNode = null;
                    
                    Criteria crit = (Criteria)node.getProperty(NodeConstants.Info.PROCEDURE_CRITERIA);
                    
                    if (crit != null) {
                        List references = (List)node.getProperty(NodeConstants.Info.PROCEDURE_INPUTS);
                        List defaults = (List)node.getProperty(NodeConstants.Info.PROCEDURE_DEFAULTS);
                        
                        peNode = new DependentProcedureExecutionNode(getID(), crit, references, defaults);                        
                    } else {
                        peNode = new PlanExecutionNode(getID());
                    }
                    
                    peNode.setProcessorPlan(plan);
                    processNode = peNode;

                } else {
                    AccessNode aNode = null;
                    Command command = (Command) node.getProperty(NodeConstants.Info.ATOMIC_REQUEST);
                    Object modelID = node.getProperty(NodeConstants.Info.MODEL_ID);
                    EvaluatableVisitor ev = null;
                    if(node.hasBooleanProperty(NodeConstants.Info.IS_DEPENDENT_SET)) {
                        if (command instanceof StoredProcedure) {
                            List references = (List)node.getProperty(NodeConstants.Info.PROCEDURE_INPUTS);
                            List defaults = (List)node.getProperty(NodeConstants.Info.PROCEDURE_DEFAULTS);
                            Criteria crit = (Criteria)node.getProperty(NodeConstants.Info.PROCEDURE_CRITERIA);
                            
                            DependentProcedureAccessNode depAccessNode = new DependentProcedureAccessNode(getID(), crit, references, defaults);
                            processNode = depAccessNode;
                            aNode = depAccessNode;
                        } else {
                            //create dependent access node
                            DependentAccessNode depAccessNode = new DependentAccessNode(getID());
                            
                            if(modelID != null){
                            	depAccessNode.setPushdown(CapabilitiesUtil.supports(Capability.DEPENDENT_JOIN, modelID, metadata, capFinder));
                                depAccessNode.setMaxSetSize(CapabilitiesUtil.getMaxInCriteriaSize(modelID, metadata, capFinder));
                                depAccessNode.setMaxPredicates(CapabilitiesUtil.getMaxDependentPredicates(modelID, metadata, capFinder));
	                            depAccessNode.setUseBindings(CapabilitiesUtil.supports(Capability.DEPENDENT_JOIN_BINDINGS, modelID, metadata, capFinder));
	                            //TODO: allow the translator to drive this property
	                            //simplistic check of whether this query is complex to re-execute
	                            Query query = (Query)command;
	                            if (query.getGroupBy() != null
	                                            || query.getFrom().getClauses().size() > 1
	                                            || !(query.getFrom().getClauses().get(0) instanceof UnaryFromClause)
	                                            || query.getWith() != null) {
		                            depAccessNode.setComplexQuery(true);
	                            } else {
		                            //check to see if there in an index on at least one of the dependent sets
		                            Set<GroupSymbol> groups = new HashSet<GroupSymbol>(query.getFrom().getGroups());
		                            boolean found = false;
		                            for (Criteria crit : Criteria.separateCriteriaByAnd(query.getCriteria())) {
			                            if (crit instanceof DependentSetCriteria) {
				                            DependentSetCriteria dsc = (DependentSetCriteria)crit;
				                            if (NewCalculateCostUtil.getKeyUsed(ElementCollectorVisitor.getElements(dsc.getExpression(), true), groups, metadata, null) != null) {
				                            	found = true;
					                            break;
				                            }
			                            }
		                            }
		                            if (!found) {
			                            depAccessNode.setComplexQuery(true);
		                            }
	                            }
                            }
                            processNode = depAccessNode;
                            aNode = depAccessNode;
                        }
                        aNode.setShouldEvaluateExpressions(true);
                    } else {

	                    // create access node
	                    aNode = new AccessNode(getID());
	                    processNode = aNode;

	                    ev = EvaluatableVisitor.needsEvaluation(command, modelID, metadata, capFinder);
	                    aNode.setShouldEvaluateExpressions(ev.requiresEvaluation(EvaluationLevel.PROCESSING));
                    }
                    //-- special handling for temp tables. currently they cannot perform projection
                    try {
                        if (command instanceof Query) {
                            processNode = correctProjectionInternalTables(node, aNode);
                        }
                    } catch (QueryMetadataException err) {
                         throw new TeiidComponentException(QueryPlugin.Event.TEIID30248, err);
                    }
					// EXTENSION
					// ISSUE: DV-99
					// FIXED BY: Salvatore Raunich
					// Applied solution from TEIID-2344
					// DATE: 2013.11.08
					if (modelID != null && !capFinder.isValid(metadata.getFullName(modelID))) {
						// TODO: we ideally want to handle the partial resutls case here differently
						// by adding a null node / and a source warning
						// for now it's just as easy to say that the user needs to take steps to
						// return static capabilities
						throw new QueryPlannerException(QueryPlugin.Event.TEIID30498, QueryPlugin.Util.gs(QueryPlugin.Event.TEIID30498,
								metadata.getFullName(modelID)));
					}
					// END EXTENSION
                    try {
                    	String modelName = metadata.getFullName(modelID);
                    	SourceCapabilities caps = capFinder.findCapabilities(modelName);
                    	if (caps.getSourceProperty(Capability.OEVRRIDE_FETCH_SIZE) != null)
                    		aNode.setOverrideFetchSize((Integer) caps.getSourceProperty(Capability.OEVRRIDE_FETCH_SIZE));
                    } catch (NullPointerException e) {
                    	// for temp tables metadata.getFullName() may throw NPE
                    }
		            setRoutingName(aNode, node, command);
					if (command instanceof Insert) {
						Insert insert = (Insert) command;
						if (insert.getQueryExpression() != null) {
							boolean selInto = (insert instanceof SelectInto);
							insert.setQueryExpression((QueryCommand) aliasCommand(aNode, insert.getQueryExpression(), modelID, selInto));
						}
					} else if (command instanceof QueryCommand) {
						command = aliasCommand(aNode, command, modelID);
					}
                    // EXTENSION
                    // ISSUE: DVCORE-414
                    // FIXED BY: Roman Pedchenko
                    // DATE: 2013.05.19
                    aNode.setIsJoin((Boolean) node.getProperty(NodeConstants.Info.IS_JOIN));
                    // END EXTENSION
                    
                    // EXTENSION
                    // ISSUE: DVCORE-451
                    // FIXED BY: Salvatore Raunich
                    // DATE: 2013.05.30
					if (node.getProperty(NodeConstants.Info.REC_OPT_IDS) != null) {
						aNode.setRecOptId((String) node.getProperty(NodeConstants.Info.REC_OPT_IDS));
					}
					if (node.getProperty(Info.DV_USED_MATTABLE_IDS) != null) {
						aNode.setMatTableIds((Set<String>) node.getProperty(Info.DV_USED_MATTABLE_IDS));
					}
                    if (node.getProperty(Info.DV_RECOPT) != null) {
	                	aNode.setDvRecOptIds((List<String>)DVCaller.call(163, node.getProperty(Info.DV_RECOPT)));
	                }
                    // END EXTENSION
                    aNode.setCommand(command);
	                Map<GroupSymbol, PlanNode> subPlans = (Map<GroupSymbol, PlanNode>) node.getProperty(Info.SUB_PLANS);

					if (modelID != null) {
						String fullName = metadata.getFullName(modelID);
						SourceCapabilities caps = capFinder.findCapabilities(fullName);
						aNode.setTransactionSupport((TransactionSupport) caps.getSourceProperty(Capability.TRANSACTION_SUPPORT));
					}
	                //it makes more sense to allow the multisource affect to be elevated above just access nodes
	                /*if (aNode.getModelId() != null && metadata.isMultiSource(aNode.getModelId())) {
		                VDBMetaData vdb = context.getVdb();
		                aNode.setShouldEvaluateExpressions(true); //forces a rewrite
		                aNode.setElements( (List) node.getProperty(NodeConstants.Info.OUTPUT_COLS) );
		                if (node.hasBooleanProperty(Info.IS_MULTI_SOURCE)) {
			                Expression ex = rewriteMultiSourceCommand(aNode.getCommand());
			                aNode.setConnectorBindingExpression(ex);
			                aNode.setMultiSource(true);
		                } else {
			                String sourceName = (String)node.getProperty(Info.SOURCE_NAME);
			                aNode.setConnectorBindingExpression(new Constant(sourceName));
		                }
	                } else if (subPlans == null){*/
	                if (subPlans == null) {
                    if (minimizeProject() && !aNode.isShouldEvaluate()) {
                    	aNode.minimizeProject(command);
                    }
                    //check if valid to share this with other nodes
                    if (ev != null && ev.getDeterminismLevel().compareTo(Determinism.COMMAND_DETERMINISTIC) >= 0 && command.areResultsCachable()) {
                    	checkForSharedSourceCommand(aNode, node);
                    }
                }
	                if (subPlans != null) {
		                QueryCommand qc = (QueryCommand)command;
		                if (qc.getWith() == null) {
			                qc.setWith(new ArrayList<WithQueryCommand>(subPlans.size()));
		                }
		                Map<GroupSymbol, RelationalPlan> plans = new LinkedHashMap<GroupSymbol, RelationalPlan>();
		                for (Map.Entry<GroupSymbol, PlanNode> entry : subPlans.entrySet()) {
			                RelationalPlan subPlan = convert(entry.getValue());
			                List<ElementSymbol> elems = ResolverUtil.resolveElementsInGroup(entry.getKey(), metadata);
			                subPlan.setOutputElements(elems);
			                plans.put(entry.getKey(), subPlan);
			                WithQueryCommand withQueryCommand = new WithQueryCommand(entry.getKey(), elems, null);
			                qc.getWith().add(withQueryCommand);
		                }
		                aNode.setSubPlans(plans);
	                }
                }
                break;

			case NodeConstants.Types.SELECT:

				Criteria crit = (Criteria) node.getProperty(NodeConstants.Info.SELECT_CRITERIA);
				if (!node.hasCollectionProperty(Info.OUTPUT_COLS)) {
					//the late optimization to create a dependent join from a subquery introduces
					//criteria that have no output elements set
					//TODO that should be cleaner, but the logic currently expects to be run after join implementation
					//and rerunning assign output elements seems excessive
					node.setProperty(Info.OUTPUT_COLS, node.getFirstChild().getProperty(Info.OUTPUT_COLS));
				}
				SelectNode selnode = new SelectNode(getID());
				selnode.setCriteria(crit);
				//in case the parent was a source
				selnode.setProjectedExpressions((List<Expression>) node.getProperty(NodeConstants.Info.PROJECT_COLS));
				processNode = selnode;
                
				break;

			case NodeConstants.Types.SORT:
			case NodeConstants.Types.DUP_REMOVE:
                if (node.getType() == NodeConstants.Types.DUP_REMOVE) {
	                processNode = new DupRemoveNode(getID());
                } else {
	                SortNode sortNode = new SortNode(getID());
	                OrderBy orderBy = (OrderBy) node.getProperty(NodeConstants.Info.SORT_ORDER);
	                if (orderBy != null) {
		                sortNode.setSortElements(orderBy.getOrderByItems());
	                }
	                if (node.hasBooleanProperty(NodeConstants.Info.IS_DUP_REMOVAL)) {
		                sortNode.setMode(Mode.DUP_REMOVE_SORT);
	                }

	                processNode = sortNode;
				}
				break;
			case NodeConstants.Types.GROUP:
				GroupingNode gnode = new GroupingNode(getID());
				SymbolMap groupingMap = (SymbolMap)node.getProperty(NodeConstants.Info.SYMBOL_MAP);
				gnode.setOutputMapping(groupingMap);
				gnode.setRemoveDuplicates(node.hasBooleanProperty(NodeConstants.Info.IS_DUP_REMOVAL));
				List<Expression> gCols = (List) node.getProperty(NodeConstants.Info.GROUP_COLS);
				OrderBy orderBy = (OrderBy) node.getProperty(Info.SORT_ORDER);
				if (orderBy == null) {
			        if (gCols != null) {
				        LinkedHashSet<Expression> exprs = new LinkedHashSet<Expression>();
				        for (Expression ex : gCols) {
					        exprs.add(SymbolMap.getExpression(ex));
				        }
				        orderBy = new OrderBy(RuleChooseJoinStrategy.createExpressionSymbols(new ArrayList<Expression>(exprs)));
			        }
				} else {
					HashSet<Expression> seen = new HashSet<Expression>();
			        for (int i = 0; i < gCols.size(); i++) {
			        	if (i < orderBy.getOrderByItems().size()) {
			        		OrderByItem orderByItem = orderBy.getOrderByItems().get(i);
							Expression ex = SymbolMap.getExpression(orderByItem.getSymbol());
					        if (!seen.add(ex)) {
						        continue;
					        }
				        	if (ex instanceof ElementSymbol) {
			            		ex = groupingMap.getMappedExpression((ElementSymbol) ex);
			            		orderByItem.setSymbol(new ExpressionSymbol("expr", ex)); //$NON-NLS-1$
			            	}
			        	} else {
			        		orderBy.addVariable(new ExpressionSymbol("expr", gCols.get(i)), OrderBy.ASC); //$NON-NLS-1$
			        	}
			        }
				}
				if (orderBy != null) {
			        gnode.setOrderBy(orderBy.getOrderByItems());
				}
				processNode = gnode;
				break;

			case NodeConstants.Types.SOURCE:
				Object source = node.getProperty(NodeConstants.Info.TABLE_FUNCTION);
				if (source instanceof XMLTable) {
					XMLTable xt = (XMLTable)source;
					XMLTableNode xtn = new XMLTableNode(getID());
					DVCaller.call(172, xtn, node, context);
					//we handle the projection filtering once here rather than repeating the
					//path analysis on a per plan basis
					updateGroupName(node, xt);
					// EXTENSION
					// ISSUE: DVCORE-1078
					// FIXED BY: Roman Pedchenko
					// DATE: 2014.03.26
					// ADOPTED FROM: https://issues.jboss.org/browse/TEIID-2141
					// Map elementMap = RelationalNode.createLookupMap(xt.getProjectedSymbols());
					Map<Expression, Integer> elementMap = RelationalNode.createLookupMap(xt.getProjectedSymbols());
					// END EXTENSION
			        List cols = (List) node.getProperty(NodeConstants.Info.OUTPUT_COLS);
					int[] projectionIndexes = RelationalNode.getProjectionIndexes(elementMap, cols);
					ArrayList<XMLColumn> filteredColumns = new ArrayList<XMLColumn>(projectionIndexes.length);
					for (int col : projectionIndexes) {
						filteredColumns.add(xt.getColumns().get(col));
					}
//					xt.getXQueryExpression().useDocumentProjection(filteredColumns, analysisRecord);
					xtn.setProjectedColumns(filteredColumns);
					xtn.setTable(xt);
					processNode = xtn;
					break;
				}
				// EXTENSION
				// ISSUE: DVCORE-1078
				// FIXED BY: Roman Pedchenko
				// DATE: 2014.03.26
				// ADOPTED FROM: https://issues.jboss.org/browse/TEIID-2141
				if (source instanceof ObjectTable) {
					ObjectTable ot = (ObjectTable)source;
					ObjectTableNode otn = new ObjectTableNode(getID());
					//we handle the projection filtering once here rather than repeating the
					//path analysis on a per plan basis
					updateGroupName(node, ot);
					Map<Expression, Integer> elementMap = RelationalNode.createLookupMap(ot.getProjectedSymbols());
			        List<Expression> cols = (List<Expression>) node.getProperty(NodeConstants.Info.OUTPUT_COLS);
					int[] projectionIndexes = RelationalNode.getProjectionIndexes(elementMap, cols);
					ArrayList<ObjectColumn> filteredColumns = new ArrayList<ObjectColumn>(projectionIndexes.length);
					for (int col : projectionIndexes) {
						filteredColumns.add(ot.getColumns().get(col));
					}
					otn.setProjectedColumns(filteredColumns);
					otn.setTable(ot);
					processNode = otn;
					break;
				}
				// END EXTENSION
				if (source instanceof TextTable) {
					TextTableNode ttn = new TextTableNode(getID());
					TextTable tt = (TextTable)source;
					updateGroupName(node, tt);
					ttn.setTable(tt);
					processNode = ttn;
					break;
				}
				if (source instanceof ArrayTable) {
					ArrayTableNode atn = new ArrayTableNode(getID());
					ArrayTable at = (ArrayTable)source;
					updateGroupName(node, at);
					atn.setTable(at);
					processNode = atn;
					break;
				}
			    SymbolMap symbolMap = (SymbolMap) node.getProperty(NodeConstants.Info.SYMBOL_MAP);
				if(symbolMap != null) {
					PlanNode child = node.getLastChild();

                	if (child.getType() == NodeConstants.Types.PROJECT
                			|| child.getType() == NodeConstants.Types.SELECT) {
                		//update the project cols based upon the original output
                		child.setProperty(NodeConstants.Info.PROJECT_COLS, child.getProperty(NodeConstants.Info.OUTPUT_COLS));
                	}
					if (child.getType() != NodeConstants.Types.SET_OP || child.getProperty(Info.SET_OPERATION) == Operation.UNION) {
						child.setProperty(NodeConstants.Info.OUTPUT_COLS, node.getProperty(NodeConstants.Info.OUTPUT_COLS));
					} else {
						//else we cannot directly update the child properties as the child will get converted to a join
						//create a projection instead for initialization purposes, but won't impact performance
						ProjectNode pNode = new ProjectNode(getID());
						pNode.setSelectSymbols((List<? extends Expression>) child.getProperty(NodeConstants.Info.OUTPUT_COLS));
						return prepareToAdd(node, pNode);
					}
				}
				return null;
    		case NodeConstants.Types.SET_OP:
                Operation setOp = (Operation) node.getProperty(NodeConstants.Info.SET_OPERATION);
                boolean useAll = ((Boolean) node.getProperty(NodeConstants.Info.USE_ALL)).booleanValue();
                if(setOp == Operation.UNION) {
                    RelationalNode unionAllNode = new UnionAllNode(getID());

                    if(useAll) {
                        processNode = unionAllNode;
                    } else {
                    	boolean onlyDupRemoval = node.hasBooleanProperty(NodeConstants.Info.IS_DUP_REMOVAL);
	                    if (onlyDupRemoval) {
		                    processNode = new DupRemoveNode(getID());
	                    } else {
		                    SortNode sNode = new SortNode(getID());
		                    sNode.setMode(Mode.DUP_REMOVE_SORT);
		                    processNode = sNode;
	                    }
                        
                        unionAllNode.setElements( (List) node.getProperty(NodeConstants.Info.OUTPUT_COLS) );
                        processNode.addChild(unionAllNode);
                    }
                } else {
                    JoinNode joinAsSet = new JoinNode(getID());
                    joinAsSet.setJoinStrategy(new MergeJoinStrategy(SortOption.SORT_DISTINCT, SortOption.SORT_DISTINCT, true));
	                //If we push these sorts, we will have to enforce null order, since nulls are equal here
                    List leftExpressions = (List) node.getFirstChild().getProperty(NodeConstants.Info.OUTPUT_COLS);
                    List rightExpressions = (List) node.getLastChild().getProperty(NodeConstants.Info.OUTPUT_COLS);
                    joinAsSet.setJoinType(setOp == Operation.EXCEPT ? JoinType.JOIN_ANTI_SEMI : JoinType.JOIN_SEMI);
                    joinAsSet.setJoinExpressions(leftExpressions, rightExpressions);
                    processNode = joinAsSet;
                }

                break;

            case NodeConstants.Types.TUPLE_LIMIT:
                Expression rowLimit = (Expression)node.getProperty(NodeConstants.Info.MAX_TUPLE_LIMIT);
                Expression offset = (Expression)node.getProperty(NodeConstants.Info.OFFSET_TUPLE_COUNT);
                LimitNode ln = new LimitNode(getID(), rowLimit, offset);
                ln.setImplicit(node.hasBooleanProperty(Info.IS_IMPLICIT_LIMIT));
                processNode = ln;
                break;

			case NodeConstants.Types.NULL:
				processNode = new NullNode(getID());
				break;

			case NodeConstants.Types.UNPIVOT:
				final ElementSymbol col = (ElementSymbol) node.getProperty(Info.UNPIVOT_COL);
				final ElementSymbol forCol = (ElementSymbol) node.getProperty(Info.UNPIVOT_FOR_COL);
				@SuppressWarnings("unchecked")
				final List<ElementSymbol> innerProjection = (List<ElementSymbol>) node.getProperty(Info.UNPIVOT_INNER_PROJECTION);
				@SuppressWarnings("unchecked")
				final Map<Integer, String> foldColumns = (Map<Integer, String>) node.getProperty(Info.UNPIVOT_FOLD_COLS);
				processNode = new UnpivotNode(getID(), foldColumns, col, forCol, innerProjection);
				break;

			default:
                 throw new QueryPlannerException(QueryPlugin.Event.TEIID30250, QueryPlugin.Util.gs(QueryPlugin.Event.TEIID30250, NodeConstants.getNodeTypeString(node.getType())));
		}

		if(processNode != null) {
			processNode = prepareToAdd(node, processNode);
		}

		return processNode;
	}
	
	private Command aliasCommand(AccessNode aNode, Command command,
			Object modelID, boolean selectInto) throws TeiidComponentException,
			QueryPlannerException {
		try {
            command = (Command)command.clone();
            boolean aliasGroups = modelID != null && (CapabilitiesUtil.supportsGroupAliases(modelID, metadata, capFinder) 
            		|| CapabilitiesUtil.supports(Capability.QUERY_FROM_INLINE_VIEWS, modelID, metadata, capFinder));
            boolean aliasColumns = modelID != null && (CapabilitiesUtil.supports(Capability.QUERY_SELECT_EXPRESSION, modelID, metadata, capFinder)
            		|| CapabilitiesUtil.supports(Capability.QUERY_FROM_INLINE_VIEWS, modelID, metadata, capFinder));
        	// EXTENSION
        	// ISSUE: DVCORE-980,DV-59
        	// FIXED BY: Roman Pedchenko
        	// DATE: 2014.02.07
//            command.acceptVisitor(new AliasGenerator(aliasGroups, !aliasColumns));
            boolean skipAliasOrderby = modelID != null && !CapabilitiesUtil.supports(Capability.QUERY_ORDERBY_ALIAS, modelID, metadata, capFinder);
            AliasGenerator visitor = new AliasGenerator(aliasGroups, !aliasColumns, skipAliasOrderby, selectInto);
			SourceHint sh = command.getSourceHint();
			if (sh != null && aliasGroups) {
				VDBMetaData vdb = context.getDQPWorkContext().getVDB();
				ModelMetaData model = vdb.getModel(aNode.getModelName());
				List<String> sourceNames = model.getSourceNames();
				SpecificHint sp = null;
				if (sourceNames.size() == 1) {
					sp = sh.getSpecificHint(sourceNames.get(0));
				}
				if (sh.isUseAliases() || (sp != null && sp.isUseAliases())) {
					visitor.setAliasMapping(context.getAliasMapping());
				}
			}
            List<Reference> references = ReferenceCollectorVisitor.getReferences(command);
            if (!references.isEmpty()) {
                 Set<String> correleatedGroups = new TreeSet<String>(StringUtil.CASE_INSENSITIVE_ORDER);
                 for (Reference ref : references) {
                         if (ref.isCorrelated() && ref.getExpression().getGroupSymbol() != null) {
                                 correleatedGroups.add(ref.getExpression().getGroupSymbol().getName());
                         }
                 }
                 visitor.setCorrelationGroups(correleatedGroups);
            }
            command.acceptVisitor(visitor);
        	// END EXTENSION
        } catch (QueryMetadataException err) {
             throw new TeiidComponentException(QueryPlugin.Event.TEIID30249, err);
		} catch (TeiidRuntimeException e) {
			if (e.getCause() instanceof QueryPlannerException) {
				throw (QueryPlannerException)e.getCause();
			}
			throw e;
		}
		return command;		
	}
	
	private Command aliasCommand(AccessNode aNode, Command command,
			Object modelID) throws TeiidComponentException,
			QueryPlannerException {
		return aliasCommand(aNode, command, modelID, false);
	}
	
	private void checkForSharedSourceCommand(AccessNode aNode, PlanNode node) {
		//create a top level key to avoid the full command toString
		String modelName = aNode.getModelName();
		Command cmd = aNode.getCommand();
		
		//don't share full scans against internal sources, it's a waste of buffering
		if (CoreConstants.SYSTEM_MODEL.equals(modelName)
				|| CoreConstants.SYSTEM_ADMIN_MODEL.equals(modelName) 
				|| CoreConstants.SYSTEM_ADMIN_DYN_MODEL.equals(modelName) 
				|| TempMetadataAdapter.TEMP_MODEL.getName().equals(modelName)) {
			if (!(cmd instanceof Query)) {
				return;
			}
			Query query = (Query)cmd;
			List<GroupSymbol> groups = query.getFrom().getGroups();
			boolean cacheable = false;
			for (GroupSymbol group : groups) {
				if (group.getMetadataID() instanceof Table) {
					cacheable = ((Table)group.getMetadataID()).isCacheable();
					if (cacheable) break;
				}
			}
			if (!cacheable && query.getOrderBy() == null && query.getCriteria() == null) {
				return;
			}
		}

		AccessNode other = sharedCommands.get(cmd);

		//lateral may be reused any number of times,
		//so this requires special handling
		//we will clean it up at the end
		boolean lateral = false;
		while (node.getParent() != null) {
			if (node.getParent().getType() == NodeConstants.Types.JOIN 
					&& node.getParent().getProperty(Info.JOIN_STRATEGY) == JoinStrategyType.NESTED_TABLE
					&& node.getParent().getLastChild() == node) {
				lateral = true;
				break;
			}
			node = node.getParent();
		}

		if (other == null) {
			sharedCommands.put(cmd, aNode);
			if (lateral) {
				aNode.info = new RegisterRequestParameter.SharedAccessInfo();
				aNode.info.id = sharedId.getAndIncrement();
				aNode.info.sharingCount = -1;
			}
		} else {
			if (other.info == null) {
				other.info = new RegisterRequestParameter.SharedAccessInfo();
				other.info.id = sharedId.getAndIncrement();
			}
			if (other.info.sharingCount != -1) {
				other.info.sharingCount++;
			} else if (lateral) {
				other.info.sharingCount = -1;
			}
            aNode.info = other.info;
		}
	}

	private void updateGroupName(PlanNode node, TableFunctionReference tt) {
		String groupName = node.getGroups().iterator().next().getName();
		tt.getGroupSymbol().setName(groupName);
		for (ElementSymbol symbol : tt.getProjectedSymbols()) {
			symbol.setGroupSymbol(new GroupSymbol(groupName));
		}
	}

    private RelationalNode correctProjectionInternalTables(PlanNode node,
                                                                AccessNode aNode) throws QueryMetadataException,
                                                                                                       TeiidComponentException {
        if (node.getGroups().size() != 1) {
            return aNode;
        }
        GroupSymbol group = node.getGroups().iterator().next();
        Object modelId = DVCaller.call(1001, group.getMetadataID());
        if (modelId == null || !(modelId instanceof AbstractMetadataRecord))
        	modelId = metadata.getModelID(group.getMetadataID());
        if (!CoreConstants.SYSTEM_MODEL.equals(metadata.getFullName(modelId)) 
        		&& !CoreConstants.SYSTEM_ADMIN_MODEL.equals(metadata.getFullName(modelId))
        		&& !CoreConstants.SYSTEM_LOG_MODEL.equals(metadata.getFullName(modelId)) 
        		&& !CoreConstants.SYSTEM_ADMIN_VDB.equals(metadata.getFullName(modelId))) {
            return aNode;
        }
        List projectSymbols = (List) node.getProperty(NodeConstants.Info.OUTPUT_COLS);
        List<ElementSymbol> acutalColumns = ResolverUtil.resolveElementsInGroup(group, metadata);
        if (projectSymbols.equals(acutalColumns)) {
        	return aNode;
        }
        node.setProperty(NodeConstants.Info.OUTPUT_COLS, acutalColumns);
        if (node.getParent() != null && node.getParent().getType() == NodeConstants.Types.PROJECT) {
            //if the parent is already a project, just correcting the output cols is enough
            return aNode;
        }
        ProjectNode pnode = new ProjectNode(getID());
  
        pnode.setSelectSymbols(projectSymbols);
        //if the following cast fails it means that we have a dependent temp table - that is not yet possible
        aNode = (AccessNode)prepareToAdd(node, aNode);
        node.setProperty(NodeConstants.Info.OUTPUT_COLS, projectSymbols);
        pnode.addChild(aNode);
        return pnode;
    }

    private RelationalNode prepareToAdd(PlanNode node,
                                          RelationalNode processNode) {
        // Set the output elements from the plan node
        List cols = (List) node.getProperty(NodeConstants.Info.OUTPUT_COLS);

        processNode.setElements(cols);
        
        // Set the Cost Estimates
        Number estimateNodeCardinality = (Number) node.getProperty(NodeConstants.Info.EST_CARDINALITY);
        processNode.setEstimateNodeCardinality(estimateNodeCardinality);
        Number estimateNodeSetSize = (Number) node.getProperty(NodeConstants.Info.EST_SET_SIZE);
        processNode.setEstimateNodeSetSize(estimateNodeSetSize);
        Number estimateDepAccessCardinality = (Number) node.getProperty(NodeConstants.Info.EST_DEP_CARDINALITY);
        processNode.setEstimateDepAccessCardinality(estimateDepAccessCardinality);
        Number estimateDepJoinCost = (Number) node.getProperty(NodeConstants.Info.EST_DEP_JOIN_COST);
        processNode.setEstimateDepJoinCost(estimateDepJoinCost);
        Number estimateJoinCost = (Number) node.getProperty(NodeConstants.Info.EST_JOIN_COST);
        processNode.setEstimateJoinCost(estimateJoinCost);
       
        return processNode;
    }

	private void setRoutingName(AccessNode accessNode, PlanNode node, Command command)
		throws QueryPlannerException, TeiidComponentException {

		// Look up connector binding name
		try {
			Object modelID = node.getProperty(NodeConstants.Info.MODEL_ID);
			Object groupID = null;
			if(modelID == null || modelID instanceof TempMetadataID) {
				if(command instanceof StoredProcedure){
					modelID = ((StoredProcedure)command).getModelID();
					groupID = ((StoredProcedure)command).getProcedureID();
				}else{
					Collection<GroupSymbol> groups = GroupCollectorVisitor.getGroups(command, true);
					GroupSymbol group = groups.iterator().next();

					modelID = metadata.getModelID(group.getMetadataID());
					groupID = group.getMetadataID();
				}
			}
			String cbName = metadata.getFullName(modelID);
			accessNode.setModelName(cbName);
			String tempModelName = (String) DVCaller.call(1001, groupID);
			if (tempModelName != null)
				accessNode.setModelName(tempModelName);
			accessNode.setModelId(modelID);
			accessNode.setConformedTo((Set<Object>) node.getProperty(Info.CONFORMED_SOURCES));
		} catch(QueryMetadataException e) {
             throw new QueryPlannerException(QueryPlugin.Event.TEIID30251, e, QueryPlugin.Util.gs(QueryPlugin.Event.TEIID30251));
		}
	}
	
	protected boolean minimizeProject() {
		return true;
	}

}
