/*
 * JBoss, Home of Professional Open Source.
 * See the COPYRIGHT.txt file distributed with this work for information
 * regarding copyright ownership.  Some portions may be licensed
 * to Red Hat, Inc. under one or more contributor license agreements.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA.
 * Modified by
 * @author  Data Virtuality GmbH
 * @date    $Date$
 *  
 */
// https://issues.jboss.org/browse/TEIID-2626
// https://issues.jboss.org/browse/TEIID-2707
// https://issues.jboss.org/browse/TEIID-2900
// https://issues.jboss.org/browse/TEIID-2902
// https://issues.jboss.org/browse/TEIID-2457
// https://issues.jboss.org/browse/TEIID-2942
// https://github.com/teiid/teiid/commit/f840ed8283aad9cbe8d3019467a88d951d695659
// https://issues.jboss.org/browse/TEIID-2482
// https://issues.jboss.org/browse/TEIID-2346
// https://issues.jboss.org/browse/TEIID-2430
// https://issues.jboss.org/browse/TEIID-2731
// https://issues.jboss.org/browse/TEIID-2747
// https://issues.jboss.org/browse/TEIID-2946
// https://issues.jboss.org/browse/TEIID-3227
// https://issues.jboss.org/browse/TEIID-4449
package org.teiid.query.processor.relational;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.teiid.api.exception.query.ExpressionEvaluationException;
import org.teiid.api.exception.query.FunctionExecutionException;
import org.teiid.common.buffer.BlockedException;
import org.teiid.common.buffer.BufferManager;
import org.teiid.common.buffer.BufferManager.TupleSourceType;
import org.teiid.common.buffer.IndexedTupleSource;
import org.teiid.common.buffer.STree;
import org.teiid.common.buffer.STree.InsertMode;
import org.teiid.common.buffer.TupleBatch;
import org.teiid.common.buffer.TupleBuffer;
import org.teiid.common.buffer.TupleBuffer.TupleBufferTupleSource;
import org.teiid.common.buffer.TupleSource;
import org.teiid.core.TeiidComponentException;
import org.teiid.core.TeiidProcessingException;
import org.teiid.core.types.DataTypeManager;
import org.teiid.language.Literal;
import org.teiid.language.SortSpecification.NullOrdering;
import org.teiid.query.eval.Evaluator;
import org.teiid.query.function.aggregate.AggregateFunction;
import org.teiid.query.function.aggregate.FullTableFunction;
import org.teiid.query.processor.ProcessorDataManager;
import org.teiid.query.processor.relational.GroupingNode.ProjectingTupleSource;
import org.teiid.query.processor.relational.SortUtility.Mode;
import org.teiid.query.sql.LanguageObject;
import org.teiid.query.sql.lang.OrderBy;
import org.teiid.query.sql.lang.OrderByItem;
import org.teiid.query.sql.symbol.AggregateSymbol.Type;
import org.teiid.query.sql.symbol.Constant;
import org.teiid.query.sql.symbol.ElementSymbol;
import org.teiid.query.sql.symbol.Expression;
import org.teiid.query.sql.symbol.WindowFunction;
import org.teiid.query.sql.symbol.WindowSpecification;
import org.teiid.query.sql.symbol.WindowSpecification.Frame;
import org.teiid.query.sql.symbol.WindowSpecification.FrameType;
import org.teiid.query.sql.util.SymbolMap;
import org.teiid.query.util.CommandContext;


public class WindowFunctionProjectNode extends SubqueryAwareRelationalNode {
	
	private static final List<Integer> SINGLE_VALUE_ID = Arrays.asList(0);

	private enum Phase {
		COLLECT,
		PROCESS,
		OUTPUT
	}
	
	private static class WindowFunctionInfo {
		WindowFunction function;
		int outputIndex;
	}
	
	private static class WindowSpecificationInfo {
		List<Integer> groupIndexes = new ArrayList<Integer>();
		List<Integer> sortIndexes = new ArrayList<Integer>();
		List<NullOrdering> nullOrderings = new ArrayList<NullOrdering>();
		List<Boolean> orderType = new ArrayList<Boolean>();
		List<WindowFunctionInfo> functions = new ArrayList<WindowFunctionInfo>();
		List<WindowFunctionInfo> rowValuefunctions = new ArrayList<WindowFunctionInfo>();
		List<WindowFunctionInfo> frameFunctions = new ArrayList<WindowFunctionInfo>();
		List<WindowFunctionInfo> partitionedFunctions = new ArrayList<WindowFunctionInfo>();
	}
	
	private LinkedHashMap<WindowSpecification, WindowSpecificationInfo> windows = new LinkedHashMap<WindowSpecification, WindowSpecificationInfo>();
	private LinkedHashMap<Expression, Integer> expressionIndexes;
	private List<int[]> passThrough = new ArrayList<int[]>();
	
	private Map<Expression, Integer> elementMap;
	
	//processing state
	private Phase phase = Phase.COLLECT;
	private TupleBuffer tb;
	private TupleSource inputTs;
	private STree[] partitionMapping;
	private STree[] valueMapping;
	private STree[] rowValueMapping;
	private STree[] frameValueMapping;
	private STree[] partValueMapping;
	private IndexedTupleSource outputTs;
	
	public WindowFunctionProjectNode(int nodeId) {
		super(nodeId);
	}
	
	protected WindowFunctionProjectNode() {
	}
	
	@Override
	public void reset() {
		super.reset();
		this.tb = null;
		this.inputTs = null;
		this.phase = Phase.COLLECT;
		this.partitionMapping = null;
		this.valueMapping = null;
		this.rowValueMapping = null;
		this.frameValueMapping = null;
		this.partValueMapping = null;
		this.outputTs = null;
	}
	
	@Override
	public void closeDirect() {
		if (tb != null) {
			tb.remove();
			tb = null;
		}
		removeMappings(partitionMapping);
		partitionMapping = null;
		removeMappings(valueMapping);
		valueMapping = null;
		removeMappings(rowValueMapping);
		rowValueMapping = null;
		removeMappings(frameValueMapping);
		frameValueMapping = null;
		removeMappings(partValueMapping);
		partValueMapping = null;
	}

	private void removeMappings(STree[] mappings) {
		if (mappings != null) {
			for (STree tree : mappings) {
				if (tree != null) {
					tree.remove();
				}
			}
		}
	}
	
	public Object clone(){
		WindowFunctionProjectNode clonedNode = new WindowFunctionProjectNode();
        this.copyTo(clonedNode);
        clonedNode.windows = windows;
        clonedNode.expressionIndexes = expressionIndexes;
        clonedNode.passThrough = passThrough;
		return clonedNode;
	}
	
	/**
	 * This state can be determined prior to initialize and is the same for all nodes,
	 * so it is moved into it's own init routine
	 */
	public void init() {
		expressionIndexes = new LinkedHashMap<Expression, Integer>();
		for (int i = 0; i < getElements().size(); i++) {
			Expression ex = SymbolMap.getExpression(getElements().get(i));
			if (ex instanceof WindowFunction) {
				WindowFunction wf = (WindowFunction)ex;
				WindowSpecification ws = wf.getWindowSpecification();
				WindowSpecificationInfo wsi = windows.get(ws);
				if (wsi == null) {
					wsi = new WindowSpecificationInfo();
					windows.put(wf.getWindowSpecification(), wsi);
					if (ws.getPartition() != null) {
						for (Expression ex1 : ws.getPartition()) {
							Integer index = GroupingNode.getIndex(ex1, expressionIndexes);
							wsi.groupIndexes.add(index);
							wsi.orderType.add(OrderBy.ASC);
							wsi.nullOrderings.add(null);
						}
					}
					if (ws.getOrderBy() != null) {
						for (OrderByItem item : ws.getOrderBy().getOrderByItems()) {
							Expression ex1 = SymbolMap.getExpression(item.getSymbol());
							Integer index = GroupingNode.getIndex(ex1, expressionIndexes);
							wsi.sortIndexes.add(index);
							wsi.orderType.add(item.isAscending());
							wsi.nullOrderings.add(item.getNullOrdering());
						}
					}
				}
				WindowFunctionInfo wfi = new WindowFunctionInfo();
				wfi.function = wf;
				//collect the agg expressions
				for (Expression e : wf.getFunction().getArgs()) {
					GroupingNode.getIndex(e, expressionIndexes);
				}
				if (wf.getFunction().getOrderBy() != null) {
					for (OrderByItem item : wf.getFunction().getOrderBy().getOrderByItems()) {
						GroupingNode.getIndex(item.getSymbol(), expressionIndexes);
					}
				}
				if (wf.getFunction().getCondition() != null) {
					GroupingNode.getIndex(wf.getFunction().getCondition(), expressionIndexes);
				}
				wfi.outputIndex = i;
				if (wf.getWindowSpecification().getFrame() != null &&
						wf.getFunction().getAggregateFunction() != Type.ROW_NUMBER &&
						wf.getFunction().getAggregateFunction() != Type.RANK &&
						wf.getFunction().getAggregateFunction() != Type.DENSE_RANK)
					wsi.frameFunctions.add(wfi);
				else
				if (wf.getFunction().getAggregateFunction() == Type.ROW_NUMBER) {
					wsi.rowValuefunctions.add(wfi);
				} else if (wf.getFunction().getAggregateFunction() == Type.LAG
						|| wf.getFunction().getAggregateFunction() == Type.LEAD
						) {
					wsi.partitionedFunctions.add(wfi);
				} else {
					wsi.functions.add(wfi);
				}
			} else {
				int index = GroupingNode.getIndex(ex, expressionIndexes);
				passThrough.add(new int[] {i, index});
			}
		}
	}

	@Override
	protected TupleBatch nextBatchDirect() throws BlockedException,
			TeiidComponentException, TeiidProcessingException {
		
		if (phase == Phase.COLLECT) {
			saveInput();
			phase = Phase.PROCESS;
			partitionMapping = new STree[this.windows.size()];
			valueMapping = new STree[this.windows.size()];
			rowValueMapping = new STree[this.windows.size()];
			frameValueMapping = new STree[this.windows.size()];
			partValueMapping = new STree[this.windows.size()];
		}
		
		if (phase == Phase.PROCESS) {
			buildResults();
			phase = Phase.OUTPUT;
		}
		
		if (phase == Phase.OUTPUT) {
			if (outputTs == null) {
				outputTs = tb.createIndexedTupleSource(true);
			}
			while (outputTs.hasNext()) {
				List<?> tuple = outputTs.nextTuple();
				int rowId = (Integer)tuple.get(tuple.size() - 1);
				int size = getElements().size();
				ArrayList<Object> outputRow = new ArrayList<Object>(size);
				for (int i = 0; i < size; i++) {
					outputRow.add(null);
				}
				for (int[] entry : passThrough) {
					outputRow.set(entry[0], tuple.get(entry[1]));
				}
				List<Map.Entry<WindowSpecification, WindowSpecificationInfo>> specs = new ArrayList<Map.Entry<WindowSpecification,WindowSpecificationInfo>>(windows.entrySet());
				for (int specIndex = 0; specIndex < specs.size(); specIndex++) {
					Map.Entry<WindowSpecification, WindowSpecificationInfo> entry = specs.get(specIndex);
					List<?> idRow = Arrays.asList(rowId);
					List<WindowFunctionInfo> functions = entry.getValue().rowValuefunctions;
					if (!functions.isEmpty()) {
						List<?> valueRow = rowValueMapping[specIndex].find(idRow);
						for (int i = 0; i < functions.size(); i++) {
							WindowFunctionInfo wfi = functions.get(i);
							outputRow.set(wfi.outputIndex, valueRow.get(i+1));
						}
					}
					functions = entry.getValue().partitionedFunctions;
					if (!functions.isEmpty()) {
						List<?> valueRow = partValueMapping[specIndex].find(idRow);
						for (int i = 0; i < functions.size(); i++) {
							WindowFunctionInfo wfi = functions.get(i);
							outputRow.set(wfi.outputIndex, valueRow.get(i+1));
						}
					}
					functions = entry.getValue().functions;
					if (!functions.isEmpty()) {
						if (partitionMapping[specIndex] != null) {
							idRow = partitionMapping[specIndex].find(idRow);
							idRow = idRow.subList(1, 2);
						} else {
							idRow = SINGLE_VALUE_ID;
						}
						List<?> valueRow = valueMapping[specIndex].find(idRow);
						for (int i = 0; i < functions.size(); i++) {
							WindowFunctionInfo wfi = functions.get(i);
							outputRow.set(wfi.outputIndex, valueRow.get(i+1));
						}
					}
					functions = entry.getValue().frameFunctions;
					if (!functions.isEmpty()) {
						List<?> valueRow = frameValueMapping[specIndex].find(idRow);
						for (int i = 0; i < functions.size(); i++) {
							WindowFunctionInfo wfi = functions.get(i);
							outputRow.set(wfi.outputIndex, valueRow.get(i + 2));
						}
					}
				}
				this.addBatchRow(outputRow);
				if (this.isBatchFull()) {
					return pullBatch();
				}
			}
			terminateBatches();
		}
		return this.pullBatch();
	}

	/**
	 * Build the results by maintaining indexes that either map
	 * rowid->values
	 * or
	 * rowid->partitionid and partitionid->values
	 * 
	 * TODO use the size hint for tree balancing
	 */
	private void buildResults() throws TeiidComponentException,
			TeiidProcessingException, FunctionExecutionException,
			ExpressionEvaluationException {
		List<Map.Entry<WindowSpecification, WindowSpecificationInfo>> specs = new ArrayList<Map.Entry<WindowSpecification,WindowSpecificationInfo>>(windows.entrySet());
		for (int specIndex = 0; specIndex < specs.size(); specIndex++) {
			Map.Entry<WindowSpecification, WindowSpecificationInfo> entry = specs.get(specIndex);
			WindowSpecificationInfo info = entry.getValue();
			IndexedTupleSource specificationTs = tb.createIndexedTupleSource();
			boolean multiGroup = false;
			int[] partitionIndexes = null;
			int[] orderIndexes = null;

			TupleBuffer sorted = null;
			//if there is partitioning or ordering, then sort
			if (!info.orderType.isEmpty()) {
				multiGroup = true;
				int[] sortKeys = new int[info.orderType.size()];
				int i = 0;
				if (!info.groupIndexes.isEmpty()) {
					for (Integer sortIndex : info.groupIndexes) {
						sortKeys[i++] = sortIndex;
					}
					partitionIndexes = Arrays.copyOf(sortKeys, info.groupIndexes.size());
				}
				if (!info.sortIndexes.isEmpty()) {
					for (Integer sortIndex : info.sortIndexes) {
						sortKeys[i++] = sortIndex;
					}
					orderIndexes = Arrays.copyOfRange(sortKeys, info.groupIndexes.size(), info.groupIndexes.size() + info.sortIndexes.size());
				}
				ElementSymbol key = new ElementSymbol("rowId"); //$NON-NLS-1$
				key.setType(DataTypeManager.DefaultDataClasses.INTEGER);
				ElementSymbol value = new ElementSymbol("partitionId"); //$NON-NLS-1$
				value.setType(DataTypeManager.DefaultDataClasses.INTEGER);
				List<ElementSymbol> elements = Arrays.asList(key, value);
				if (!info.functions.isEmpty())
					partitionMapping[specIndex] = this.getBufferManager().createSTree(elements, this.getConnectionID(), 1);
				SortUtility su = new SortUtility(null, Mode.SORT, this.getBufferManager(), this.getConnectionID(), tb.getSchema(), info.orderType, info.nullOrderings, sortKeys);
				su.setWorkingBuffer(tb);
				su.setNonBlocking(true);
				sorted = su.sort();
				specificationTs = sorted.createIndexedTupleSource();
			}
			List<AggregateFunction> aggs = initializeAccumulators(info.functions, specIndex, ValueMapping.VALUE);
			List<AggregateFunction> rowValueAggs = initializeAccumulators(info.rowValuefunctions, specIndex, ValueMapping.ROW);
			List<AggregateFunction> partAggs = initializeAccumulators(info.partitionedFunctions, specIndex, ValueMapping.PARTITIONED);

		    List<AggregateFunction> frameAggs = initializeFrameAccumulators(frameValueMapping, info.frameFunctions, specIndex);
    		int aggCount = frameAggs.size();

		    int groupId = 0;
			List<?> lastRow = null;
	    	int startRow = specificationTs.getCurrentIndex();
	    	int startPartition = startRow;
	    	int endRow = tb.getRowCount() - startRow + 1;
			TupleBufferTupleSource secondaryTs = null;
			while (specificationTs.hasNext()) {
				secondaryTs = sorted != null ? sorted.createIndexedTupleSource() : null;
    			int currentRow = specificationTs.getCurrentIndex();
				List<?> tuple = specificationTs.nextTuple();

				if (multiGroup) {
				    if (lastRow != null) {
				    	boolean samePartition = GroupingNode.sameGroup(partitionIndexes, tuple, lastRow);
				    	if (!aggs.isEmpty() && (!samePartition || !GroupingNode.sameGroup(orderIndexes, tuple, lastRow))) {
			        		saveValues(specIndex, aggs, groupId, samePartition, ValueMapping.VALUE);
		        			groupId++;
				    	}
				    	if (secondaryTs != null && !partAggs.isEmpty() && !samePartition) {
				    		countPartFunctions(specIndex, partAggs, startPartition, currentRow - 1, secondaryTs);
					    	startPartition = currentRow;
				    	}
		        		saveValues(specIndex, rowValueAggs, lastRow.get(lastRow.size() - 1), samePartition, ValueMapping.ROW);
		        	}
				    if (!aggs.isEmpty()) {
			        	List<Object> partitionTuple = Arrays.asList(tuple.get(tuple.size() - 1), groupId);
						partitionMapping[specIndex].insert(partitionTuple, InsertMode.NEW, -1);
				    }
		        }
		        for (AggregateFunction function : aggs) {
		        	function.addInput(tuple, getContext());
		        }
		        for (AggregateFunction function : rowValueAggs) {
		        	function.addInput(tuple, getContext());
		        }
			    if (!frameAggs.isEmpty() && sorted != null) {
	    			Integer rowId = (Integer) tuple.get(tuple.size() - 1);
			    	IndexedTupleSource tempTs = sorted.createIndexedTupleSource();
			    	List<FrameFunctionContext> ffCtxs = new ArrayList<WindowFunctionProjectNode.FrameFunctionContext>();
		    		int i = 0;
			    	for (AggregateFunction function : frameAggs) {
			    		ffCtxs.add(new FrameFunctionContext(entry.getKey().getFrame(), function, i, tuple, currentRow));
		    			i++;
			    	}

			    	tempTs.reset();
			    	while (tempTs.hasNext()) {
			    		int intCurrentIndex = tempTs.getCurrentIndex();
			    		List<?> intTuple = tempTs.nextTuple();
			    		if (!GroupingNode.sameGroup(partitionIndexes, intTuple, tuple))
			    			continue;
			    		for (FrameFunctionContext ffCtx : ffCtxs)
							processFrame(ffCtx, intTuple, intCurrentIndex);
			    	}
			    	for (FrameFunctionContext ffCtx : ffCtxs) {
			    		saveFrameValue(frameValueMapping, InsertMode.NEW, specIndex, ffCtx.function.getResult(getContext()), ffCtx.funcIndex, aggCount, rowId, currentRow);
			    		ffCtx.function.reset();
			    	}
			    }
		        lastRow = tuple;
			}
		    if(lastRow != null) {
		    	saveValues(specIndex, aggs, groupId, true, ValueMapping.VALUE);
	    		countPartFunctions(specIndex, partAggs, startPartition, endRow, secondaryTs);
		    	saveValues(specIndex, rowValueAggs, lastRow.get(lastRow.size() - 1), true, ValueMapping.ROW);
		    }
		}
	}

	private void countPartFunctions(int specIndex, List<AggregateFunction> partAggs, int partStartRow, int partEndRow, TupleBufferTupleSource secondaryTs)
			throws TeiidProcessingException, TeiidComponentException, FunctionExecutionException, ExpressionEvaluationException {
		if (partAggs.isEmpty())
			return;
		for (int partCurrentRow = partStartRow; partCurrentRow <= partEndRow; partCurrentRow++) {
			secondaryTs.setPosition(partCurrentRow);
			List<?> currentRowData = secondaryTs.nextTuple();
			Object id = currentRowData.get(currentRowData.size() - 1);
			for (AggregateFunction function : partAggs) {
				if (function instanceof FullTableFunction) {
					((FullTableFunction) function).addInputTable(secondaryTs, partCurrentRow, partStartRow, partEndRow, getContext());
				}
			}
			saveValues(specIndex, partAggs, id, false, ValueMapping.PARTITIONED);
		}
	}

	private static final class FrameFunctionContext {
		private Frame frame;
		private AggregateFunction function;
		private int funcIndex;
		private int currentRow;

		// global
		private int valueIndex;
		private BigDecimal startExpr;
		private BigDecimal endExpr;
		private BigDecimal value;
		private FrameType startType;
		private FrameType endType;
		private boolean startFollowing;
		private boolean endFollowing;
		private boolean openUpper = false;
		private boolean started = false;
		private boolean ended = false;

		// per row
		BigDecimal value2;
		boolean fitStart;
		boolean fitEnd;

		private FrameFunctionContext(Frame frame, AggregateFunction function, int funcIndex, List<?> extTuple, int currentRow) throws TeiidProcessingException {
			this.frame = frame;
			this.function = function;
			this.funcIndex = funcIndex;
			this.currentRow = currentRow;
			valueIndex = function.getArgIndexes().length > 0 ? function.getArgIndexes()[0] : -1;
			startExpr = resolveExpr(frame.getStartExpr());
			endExpr = resolveExpr(frame.getEndExpr());
			value = calcValue(frame, valueIndex, currentRow, extTuple);
			startType = frame.getStartType();
			endType = frame.getEndType();
			if (endType == FrameType.NONE)
				endType = FrameType.ROW;
			startFollowing = frame.isStartFollowing();
			endFollowing = frame.isEndFollowing();
		}
		private BigDecimal resolveExpr(Expression expr) {
			return expr != null ? 
					new BigDecimal(((Number) (expr instanceof Literal ? ((Literal) expr).getValue() : ((Constant) expr).getValue())).doubleValue())
					: BigDecimal.ZERO;
		}
		private BigDecimal calcValue(Frame frame, int valueIndex, int currentIndex, List<?> tuple) throws TeiidProcessingException {
			return frame.isRange() && valueIndex >= 0 ? getTupleValue(tuple, valueIndex) : new BigDecimal(currentIndex);
		}
		private BigDecimal getTupleValue(List<?> extTuple, int valueIndex) throws TeiidProcessingException {
			Object value = extTuple.get(valueIndex);
			if (value == null)
				return null;
			return BigDecimal.valueOf(((Number) value).doubleValue());
		}
		private void initRow(List<?> tuple, int currentIndex) throws TeiidProcessingException {
			value2 = calcValue(frame, valueIndex, currentIndex, tuple);
			fitStart = false;
			fitEnd = false;
		}
	}
	private void processFrame(FrameFunctionContext ffCtx, List<?> tuple, int currentIndex)
			throws TeiidComponentException, TeiidProcessingException {
		ffCtx.initRow(tuple, currentIndex);
			switch (ffCtx.startType) {
			case ROW:
				if (!ffCtx.frame.isRange())
					ffCtx.fitStart = currentIndex >= ffCtx.currentRow;
				else {
					if (!ffCtx.started)
						ffCtx.started = ffCtx.value2 != null && ffCtx.value2.compareTo(ffCtx.value) == 0 || ffCtx.value2 == null && ffCtx.value == null;
					ffCtx.fitStart = ffCtx.started;
				}
				break;
			case EXPR:
				ffCtx.fitStart = ffCtx.value2 != null && ffCtx.value2.compareTo(calcRange(ffCtx.value, ffCtx.startExpr, ffCtx.startFollowing)) >= 0;
				ffCtx.openUpper = ffCtx.fitStart;
				break;
			default:
				ffCtx.fitStart = true;
				ffCtx.openUpper = true;
			}
			switch (ffCtx.endType) {
			case ROW:
				if (!ffCtx.frame.isRange())
					ffCtx.fitEnd = currentIndex <= ffCtx.currentRow;
				else {
					if (ffCtx.started || ffCtx.openUpper) {
						if (!ffCtx.ended) {
							boolean check = ffCtx.value2 != null && ffCtx.value2.compareTo(ffCtx.value) != 0 || ffCtx.value2 == null && ffCtx.value != null;
							if (!ffCtx.started)
								ffCtx.started = !check;
							else
								ffCtx.ended = check;
						}
						ffCtx.fitEnd = !ffCtx.ended;
					}
				}
				break;
			case EXPR:
				ffCtx.fitEnd = ffCtx.value2 != null && ffCtx.value2.compareTo(calcRange(ffCtx.value, ffCtx.endExpr, ffCtx.endFollowing)) <= 0;
				break;
			default:
				ffCtx.fitEnd = true;
			}
			if (ffCtx.fitStart && ffCtx.fitEnd)
				ffCtx.function.addInput(tuple, getContext());
	}
	private BigDecimal calcRange(BigDecimal value, BigDecimal expr, boolean following) {
		return following ? value.add(expr) : value.subtract(expr);
	}
	private void saveFrameValue(STree[] mapping, InsertMode mode, int specIndex, Object value, int aggIndex, int aggCount, Object id, Integer rowNo) throws FunctionExecutionException,
			ExpressionEvaluationException, TeiidComponentException, TeiidProcessingException {
		@SuppressWarnings("unchecked")
		List<Object> row = mapping[specIndex].find(Arrays.asList(id));
		if (row == null) {
			row = new ArrayList<Object>();
			row.add(id);
		} else
			mode = InsertMode.UPDATE;
		while (row.size() < aggCount + 3)
			row.add(null);
		if (value != null)
			row.set(aggIndex + 2, value);
		if (rowNo != null)
			row.set(1, rowNo);
		mapping[specIndex].insert(row, mode, -1);
	}
	private List<AggregateFunction> initializeFrameAccumulators(STree[] mapping, List<WindowFunctionInfo> frameFunctions, int specIndex) {
		List<AggregateFunction> aggs = new ArrayList<AggregateFunction>(frameFunctions.size());
		if (frameFunctions.isEmpty()) {
			return aggs;
		}
		List<ElementSymbol> elements = new ArrayList<ElementSymbol>(frameFunctions.size());
		ElementSymbol es = new ElementSymbol("rowId");
		es.setType(DataTypeManager.DefaultDataClasses.INTEGER);
		elements.add(es);
		es = new ElementSymbol("rowNo");
		es.setType(DataTypeManager.DefaultDataClasses.INTEGER);
		elements.add(es);
		for (WindowFunctionInfo wfi : frameFunctions) {
			aggs.add(GroupingNode.initAccumulator(wfi.function.getFunction(), this, expressionIndexes));
			Class<?> outputType = wfi.function.getType();
			ElementSymbol value = new ElementSymbol("val"); //$NON-NLS-1$
		    value.setType(outputType);
		    elements.add(value);
		}
		mapping[specIndex] = this.getBufferManager().createSTree(elements, this.getConnectionID(), 1);
		return aggs;
	}

	private enum ValueMapping {
		VALUE, ROW, PARTITIONED
	};

	private void saveValues(int specIndex,
			List<AggregateFunction> aggs, Object id,
			boolean samePartition, ValueMapping mapping) throws FunctionExecutionException,
			ExpressionEvaluationException, TeiidComponentException,
			TeiidProcessingException {
		if (aggs.isEmpty()) {
			return;
		}
		List<Object> row = new ArrayList<Object>(aggs.size() + 1);
		row.add(id);
		for (AggregateFunction function : aggs) {
			row.add(function.getResult(getContext()));
			if (!samePartition) {
				function.reset();
			}
		}
		switch (mapping) {
		case ROW:
			rowValueMapping[specIndex].insert(row, InsertMode.NEW, -1);
			break;
		case PARTITIONED:
			partValueMapping[specIndex].insert(row, InsertMode.NEW, -1);
			break;
		default:
			valueMapping[specIndex].insert(row, InsertMode.ORDERED, -1);	
		}
	}

	/**
	 * @param functions
	 * @param specIndex
	 * @param rowValues
	 * @return
	 */
	private List<AggregateFunction> initializeAccumulators(List<WindowFunctionInfo> functions, int specIndex, ValueMapping mapping) {
		List<AggregateFunction> aggs = new ArrayList<AggregateFunction>(functions.size());
		if (functions.isEmpty()) {
			return aggs;
		}
		List<ElementSymbol> elements = new ArrayList<ElementSymbol>(functions.size());
		ElementSymbol es = new ElementSymbol("rowId");
		es.setType(DataTypeManager.DefaultDataClasses.INTEGER);
		elements.add(es);
		for (WindowFunctionInfo wfi : functions) {
			aggs.add(GroupingNode.initAccumulator(wfi.function.getFunction(), this, expressionIndexes));
			Class<?> outputType = wfi.function.getType();
		    ElementSymbol value = new ElementSymbol("val"); //$NON-NLS-1$
		    value.setType(outputType);
		    elements.add(value);
		}
		switch (mapping) {
		case ROW:
			rowValueMapping[specIndex] = this.getBufferManager().createSTree(elements, this.getConnectionID(), 1);
			break;
		case PARTITIONED:
			partValueMapping[specIndex] = this.getBufferManager().createSTree(elements, this.getConnectionID(), 1);
			break;
		default:
			valueMapping[specIndex] = this.getBufferManager().createSTree(elements, this.getConnectionID(), 1);
		}
		return aggs;
	}

	/**
	 * Save the input generating any necessary expressions and adding a row id
	 * @param collectedExpressions
	 * @return
	 * @throws TeiidComponentException
	 * @throws TeiidProcessingException
	 */
	private void saveInput()
			throws TeiidComponentException, TeiidProcessingException {
		if (inputTs == null) {
			List<Expression> collectedExpressions = new ArrayList<Expression>(expressionIndexes.keySet());
			Evaluator eval = new Evaluator(elementMap, getDataManager(), getContext());
			final RelationalNode sourceNode = this.getChildren()[0];
			inputTs = new ProjectingTupleSource(sourceNode, eval, collectedExpressions, elementMap) {
				int index = 0;
				@Override
				public List<Object> nextTuple() throws TeiidComponentException,
						TeiidProcessingException {
					List<Object> tuple = super.nextTuple();
					if (tuple != null) {
						tuple.add(index++);
					}
					return tuple;
				}
			};
			List<ElementSymbol> schema = new ArrayList<ElementSymbol>(collectedExpressions.size() + 1);
			int index = 0;
			for (Expression ex : collectedExpressions) {
				ElementSymbol es = new ElementSymbol(String.valueOf(index++));
				es.setType(ex.getType());
				schema.add(es);
			}
			//add in the row id
			ElementSymbol es = new ElementSymbol(String.valueOf(index++));
			es.setType(DataTypeManager.DefaultDataClasses.INTEGER);
			schema.add(es);
			tb = this.getBufferManager().createTupleBuffer(schema, this.getConnectionID(), TupleSourceType.PROCESSOR);
		}
		
		List<?> tuple = null;
		while ((tuple = inputTs.nextTuple()) != null) {
			tb.addTuple(tuple);
		}
		tb.close();
		inputTs.closeSource();
		inputTs = null;
	}

	@Override
	public void initialize(CommandContext context, BufferManager bufferManager,
			ProcessorDataManager dataMgr) {
		super.initialize(context, bufferManager, dataMgr);
		if (this.elementMap == null) {
			List<? extends Expression> sourceElements = this.getChildren()[0].getElements();
			this.elementMap = createLookupMap(sourceElements);
		}
	}
	
	@Override
	public Collection<? extends LanguageObject> getObjects() {
		return getElements();
	}
}
