/*
 * JBoss, Home of Professional Open Source.
 * See the COPYRIGHT.txt file distributed with this work for information
 * regarding copyright ownership.  Some portions may be licensed
 * to Red Hat, Inc. under one or more contributor license agreements.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA.
 * Modified by
 * @author  Data Virtuality GmbH
 * @date    $Date$
 *  
 */
 
 package org.teiid.query.sql.symbol;

import java.util.List;

import org.teiid.core.util.EquivalenceUtil;
import org.teiid.core.util.HashCodeUtil;
import org.teiid.query.sql.LanguageObject;
import org.teiid.query.sql.LanguageVisitor;
import org.teiid.query.sql.lang.OrderBy;
import org.teiid.query.sql.visitor.SQLStringVisitor;

public class WindowSpecification implements LanguageObject {
	
	private List<Expression> partition;
	private OrderBy orderBy;

	public WindowSpecification() {
		
	}
	
	public List<Expression> getPartition() {
		return partition;
	}
	
	public void setPartition(List<Expression> grouping) {
		this.partition = grouping;
	}
	
	public OrderBy getOrderBy() {
		return orderBy;
	}
	
	public void setOrderBy(OrderBy orderBy) {
		this.orderBy = orderBy;
	}

	@Override
	public void acceptVisitor(LanguageVisitor visitor) {
		visitor.visit(this);
	}
	
	@Override
	public int hashCode() {
		return HashCodeUtil.hashCode(0, partition, orderBy);
	}
	
	@Override
	public boolean equals(Object obj) {
		if (obj == this) {
			return true;
		}
		if (!(obj instanceof WindowSpecification)) {
			return false;
		}
		WindowSpecification other = (WindowSpecification)obj;
		return EquivalenceUtil.areEqual(this.partition, other.partition) &&
				// EXTENSION
				// ISSUE: DV-84
				// FIXED BY: Roman Pedchenko
				// DATE: 2013.06.24
				EquivalenceUtil.areEqual(this.frame, other.frame) &&
				// END EXTENSION
		EquivalenceUtil.areEqual(this.orderBy, other.orderBy);
	}
	
	@Override
	public WindowSpecification clone() {
		WindowSpecification clone = new WindowSpecification();
		if (this.partition != null) {
			clone.setPartition(LanguageObject.Util.deepClone(this.partition, Expression.class));
		}
		if (this.orderBy != null) {
			clone.setOrderBy(this.orderBy.clone());
		}
		// EXTENSION
		// ISSUE: DV-84
		// FIXED BY: Roman Pedchenko
		// DATE: 2013.06.24
		if (this.frame != null)
			clone.setFrame(this.frame.clone());
		// END EXTENSION
		return clone;
	}
	
	@Override
	public String toString() {
		return SQLStringVisitor.getSQLString(this);
	}

	// EXTENSION
	// ISSUE: DV-84
	// FIXED BY: Roman Pedchenko
	// DATE: 2013.06.24
	private Frame frame;
	public Frame getFrame() {
		return frame;
	}

	public void setFrame(Frame frame) {
		this.frame = frame;
	}

	public static enum FrameType { NONE, UNBND, ROW, EXPR } 
	public static class Frame implements LanguageObject {
	    private boolean range;
	    private Expression startExpr;
	    private Expression endExpr;
	    private FrameType startType = FrameType.NONE;
	    private FrameType endType = FrameType.NONE;
	    private boolean startFollowing;
	    private boolean endFollowing;

		public boolean isRange() {
			return range;
		}

		public void setRange(boolean range) {
			this.range = range;
		}

		public Expression getStartExpr() {
			return startExpr;
		}

		public void setStartExpr(Expression startExpr) {
			this.startExpr = startExpr;
		}

		public Expression getEndExpr() {
			return endExpr;
		}

		public void setEndExpr(Expression endExpr) {
			this.endExpr = endExpr;
		}

		public FrameType getStartType() {
			return startType;
		}

		public void setStartType(FrameType startType) {
			this.startType = startType;
		}

		public FrameType getEndType() {
			return endType;
		}

		public void setEndType(FrameType endType) {
			this.endType = endType;
		}

		public boolean isStartFollowing() {
			return startFollowing;
		}

		public void setStartFollowing(boolean startFollowing) {
			this.startFollowing = startFollowing;
		}

		public boolean isEndFollowing() {
			return endFollowing;
		}

		public void setEndFollowing(boolean endFollowing) {
			this.endFollowing = endFollowing;
		}

		@Override
		public void acceptVisitor(LanguageVisitor visitor) {
			visitor.visit(this);
		}

		@Override
		public int hashCode() {
			return HashCodeUtil.hashCode(0, range, startExpr, endExpr, startType, endType, startFollowing, endFollowing);
		}

		@Override
		public boolean equals(Object obj) {
			if (obj == this) {
				return true;
			}
			if (!(obj instanceof Frame)) {
				return false;
			}
			Frame other = (Frame) obj;
			return this.range == other.range && this.startFollowing == other.startFollowing && this.endFollowing == other.endFollowing &&
					this.startType == other.startType && this.endType == other.endType && 
					EquivalenceUtil.areEqual(this.startExpr, other.startExpr) && EquivalenceUtil.areEqual(this.endExpr, other.endExpr);
		}

		@Override
		public Frame clone() {
			Frame clone = new Frame();
			clone.range = this.range;
			clone.startFollowing = this.startFollowing;
			clone.endFollowing = this.endFollowing;
			clone.startType = this.startType;
			clone.endType = this.endType; 
			clone.startExpr = this.startExpr;
			clone.endExpr = this.endExpr;
			return clone;
		}

		@Override
		public String toString() {
			return SQLStringVisitor.getSQLString(this);
		}
	}
	// END EXTENSION
}
