/**
 * Copyright © 2012-2022 by Data Virtuality GmbH
 */
package org.teiid.auth.openid;

import com.datavirtuality.dv.core.interfaces.DVCaller;
import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import io.jsonwebtoken.Claims;
import io.jsonwebtoken.Header;
import io.jsonwebtoken.Jws;
import io.jsonwebtoken.Jwt;
import io.jsonwebtoken.JwtParser;
import io.jsonwebtoken.Jwts;
import java.math.BigInteger;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.KeyFactory;
import java.security.interfaces.RSAPublicKey;
import java.security.spec.RSAPublicKeySpec;
import java.util.Optional;
import java.util.stream.Collectors;
import org.jboss.security.SimpleGroup;
import org.jboss.security.SimplePrincipal;
import org.teiid.adminapi.impl.OpenIDMetadata;
import org.teiid.auth.openid.model.IntrospectionResponse;
import org.teiid.auth.openid.model.JwksKey;
import org.teiid.auth.openid.model.JwksResponse;
import org.teiid.auth.openid.model.OpenIdConfigurationResponse;
import org.teiid.auth.openid.model.TokenResponse;
import org.teiid.core.util.PropertiesUtils;
import org.teiid.dqp.service.SessionServiceException;
import org.teiid.logging.LogConstants;
import org.teiid.logging.LogManager;
import org.teiid.runtime.RuntimePlugin;
import org.teiid.security.Credentials;
import org.teiid.security.SecurityHelper;
import org.teiid.security.TeiidLoginContext;

import javax.security.auth.Subject;
import java.nio.charset.StandardCharsets;
import java.security.Principal;
import java.util.Base64;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.concurrent.TimeUnit;

import static org.teiid.services.SessionServiceImpl.AT;

/**
 * @author Archil Gabeskiria, Dmitry Pogorelov
 * @date 2022.08.01
 */
public class OpenIDConnectManager {

    public static final String DV_AUTH_SERVER_URL = "dvAuthServerUrl";
    public static final String DV_AUTH_REALM = "dvAuthRealm";
    public static final String DV_AUTH_CLIENT_ID = "dvAuthClientId";
    public static final String DV_AUTH_CLIENT_SECRET = "dvAuthClientSecret";
    public static final String SUBSYSTEM_NAME = "SSO";
    public static final String DOMAIN = AT + SUBSYSTEM_NAME;

    private static final Gson GSON = new GsonBuilder().create();
    private static final long accessTokenRefreshMillis = TimeUnit.MINUTES.toMillis(
            PropertiesUtils.getIntProperty(System.getProperties(), "com.datavirtuality.openIDTokenRefreshMinutes", 1)); // when X minutes or less is left

    public TeiidLoginContext authenticate(String username, Credentials credentials, Properties properties, String securityDomain,
                                                 SecurityHelper securityHelper, OpenIDMetadata openIDMetadata) throws SessionServiceException {
        String configUrl = properties.getProperty("configurationUrl");
        String clientId = properties.getProperty("clientId");
        String clientSecret = properties.getProperty("clientSecret");
        String accessToken = properties.getProperty("accessToken");
        String refreshToken = properties.getProperty("refreshToken");
        String expiresIn = properties.getProperty("expiresIn");
        boolean useDvRoles = Boolean.valueOf(properties.getProperty("useDvRoles"));
        String provider = properties.getProperty("provider");
        String realm = properties.getProperty("realm");

        if (configUrl == null || clientId == null) {
            throw new SessionServiceException(RuntimePlugin.Event.TEIID40200, RuntimePlugin.Util.gs(RuntimePlugin.Event.TEIID40200));
        }

        // Retrieve configuration
        OpenIdConfigurationResponse config = getOpenIdConfiguration(configUrl);
        openIDMetadata.setTokenEndpoint(config.getTokenEndpoint());
        openIDMetadata.setIntrospectionEndpoint(config.getIntrospectionEndpoint());
        openIDMetadata.setUserInfoEndpoint(config.getUserInfoEndpoint());
        openIDMetadata.setJwksEndpoint(config.getJwksEndpoint());
        openIDMetadata.setClientId(clientId);
        openIDMetadata.setClientSecret(clientSecret);
        openIDMetadata.setAccessToken(accessToken);
        openIDMetadata.setRefreshToken(refreshToken);
        openIDMetadata.setUseDvRoles(useDvRoles);
        openIDMetadata.setProvider(provider);
        if (expiresIn != null && !expiresIn.isEmpty())
            openIDMetadata.setTokenExpirationTime(System.currentTimeMillis() + TimeUnit.SECONDS.toMillis(Long.valueOf(expiresIn)));

        String issuer = config.getIssuer();
        if (issuer != null && !issuer.isEmpty() && issuer.contains("/realms/")) {
            String[] issuerParts = issuer.split("/realms/");
            openIDMetadata.setServerUrl(issuerParts[0]);
            openIDMetadata.setRealm(issuerParts[1]);
        } else {
            // let's try to get serverUrl from the configUrl
            try {
                URL configURL = new URL(configUrl);
                String serverUrl = configURL.getProtocol() + "://" + configURL.getHost();
                if (configURL.getPort() > 0)
                    serverUrl = serverUrl + ":" + configURL.getPort();
                openIDMetadata.setServerUrl(serverUrl);
                openIDMetadata.setRealm(realm);
            } catch (MalformedURLException e) {
                LogManager.logError(LogConstants.CTX_RUNTIME, e, "Error parsing configUrl property");
            }
        }

        Subject openIDSubject;
        // Introspect token
        if ( (accessToken != null && !accessToken.isEmpty()) || (refreshToken != null && !refreshToken.isEmpty())) {
            LogManager.logDetail(LogConstants.CTX_RUNTIME, RuntimePlugin.Util.gs(RuntimePlugin.Event.TEIID40203, "Introspecting access token"));
            IntrospectionResponse introspection = getIntrospection(openIDMetadata, false);
            if (introspection == null) {
                openIDMetadata.setTokenExpirationTime(0);
                if (checkAndRefreshAccessToken(openIDMetadata, true)) {
                    LogManager.logDetail(LogConstants.CTX_RUNTIME, RuntimePlugin.Util.gs(RuntimePlugin.Event.TEIID40203, "Refreshing access token"));
                    introspection = getIntrospection(openIDMetadata, true);
                }

                if (introspection == null)
                    throw new SessionServiceException(RuntimePlugin.Event.TEIID40202, RuntimePlugin.Util.gs(RuntimePlugin.Event.TEIID40202));
            }

            // create subject
            openIDSubject = createSubject(openIDMetadata, introspection.getClientId(), introspection.getUsername());
            openIDMetadata.setTokenExpirationTime(TimeUnit.SECONDS.toMillis(introspection.getExp()));
        } else {
            // Retrieve token
            TokenResponse tokenResponse;
            try {
                Map<String, String> additionalParams = new HashMap<>();
                additionalParams.put("username", username);
                additionalParams.put("password", new String(credentials.getCredentialsAsCharArray()));

                tokenResponse = getToken(config.getTokenEndpoint(), "password", clientId, clientSecret, additionalParams);
                openIDMetadata.setAccessToken(tokenResponse.getAccessToken());
                openIDMetadata.setRefreshToken(tokenResponse.getRefreshToken());
            } catch (Exception e) {
                LogManager.logError(LogConstants.CTX_RUNTIME, e, "Error retrieving a token");
                throw new SessionServiceException(RuntimePlugin.Event.TEIID40201, RuntimePlugin.Util.gs(RuntimePlugin.Event.TEIID40201));
            }

            // create subject
            openIDSubject = createSubject(openIDMetadata, clientId, username);
            openIDMetadata.setTokenExpirationTime(System.currentTimeMillis() + TimeUnit.SECONDS.toMillis(tokenResponse.getExpiresIn()));
        }

        Optional<Principal> optional = openIDSubject.getPrincipals().stream().findFirst();
        if (optional.isPresent())
            username = optional.get().getName();
        return new TeiidLoginContext(username + AT + securityDomain, openIDSubject, securityDomain, securityHelper.getSecurityContext());
    }

    public static boolean checkAndRefreshAccessToken(OpenIDMetadata openIDMetadata) {
        return checkAndRefreshAccessToken(openIDMetadata, false);
    }

    private static boolean checkAndRefreshAccessToken(OpenIDMetadata openIDMetadata, boolean force) {
        long timeLeft = openIDMetadata.getTokenExpirationTime() - System.currentTimeMillis();
        if (openIDMetadata.getRefreshToken() == null || openIDMetadata.getRefreshToken().isEmpty() ||
                (!force && (timeLeft <= 0 || timeLeft > accessTokenRefreshMillis))) {
            return false;
        }
        long initialTime = System.currentTimeMillis();
        TokenResponse tokenResponse = refreshAccessToken(openIDMetadata.getTokenEndpoint(),
                openIDMetadata.getRefreshToken(), openIDMetadata.getClientId(), openIDMetadata.getClientSecret());

        openIDMetadata.setAccessToken(tokenResponse.getAccessToken());
        // every time when we refresh access token the refresh token can be updated as well
        if (tokenResponse.getRefreshToken() != null)
            openIDMetadata.setRefreshToken(tokenResponse.getRefreshToken());
        openIDMetadata.setTokenExpirationTime(initialTime + TimeUnit.SECONDS.toMillis(tokenResponse.getExpiresIn()));

        return true;
    }

    private OpenIdConfigurationResponse getOpenIdConfiguration(String configurationEndpoint) throws SessionServiceException {
        String responseJson = HttpClient.get(configurationEndpoint);
        if (responseJson == null) {
            throw new SessionServiceException(RuntimePlugin.Event.TEIID40201, "Empty response from " + configurationEndpoint);
        }
        return GSON.fromJson(responseJson, OpenIdConfigurationResponse.class);
    }

    private static TokenResponse refreshAccessToken(String tokenEndpoint, String refreshToken, String clientId, String clientSecret) {

        Map<String, String> additionalParams = new HashMap<>();
        additionalParams.put("refresh_token", refreshToken);

        return getToken(tokenEndpoint, "refresh_token", clientId, clientSecret, additionalParams);
    }

    private IntrospectionResponse getIntrospection(OpenIDMetadata openIDMetadata, boolean throwException) throws SessionServiceException {
        IntrospectionResponse introspection = introspectToken(openIDMetadata, throwException);
        if (introspection != null && Boolean.TRUE.equals(introspection.getActive()) && introspection.getExp() != null) {
            LogManager.logDetail(LogConstants.CTX_RUNTIME, RuntimePlugin.Util.gs(RuntimePlugin.Event.TEIID40203, "The access token is not active or \"exp\" claim in it is null"));
            return introspection;
        }
        return null;
    }

    private IntrospectionResponse introspectToken(OpenIDMetadata openIDMetadata, boolean throwException) throws SessionServiceException {
        IntrospectionResponse result = null;
        Map<String, String> additionalHeaders = new HashMap<>();
        String responseJson;
        if (openIDMetadata.getIntrospectionEndpoint() != null) {
            StringBuilder params = getAuthorizationParams(new StringBuilder(), openIDMetadata.getClientId(), openIDMetadata.getClientSecret());
            // client auth if needed
            /*if (openIDMetadata.getClientId() != null && openIDMetadata.getClientSecret() != null) {
                additionalHeaders.put("Authorization", basicAuth(openIDMetadata.getClientId(), openIDMetadata.getClientSecret()));
            }*/
            if (params.length() > 0)
                params.append("&");
            params.append("token=").append(openIDMetadata.getAccessToken());
            // post
            responseJson = HttpClient.post(openIDMetadata.getIntrospectionEndpoint(), params.toString(), additionalHeaders);
            if (responseJson != null)
                result = GSON.fromJson(responseJson, IntrospectionResponse.class);
            else if (throwException)
                throw new SessionServiceException(RuntimePlugin.Event.TEIID40201, "Empty response from " + openIDMetadata.getIntrospectionEndpoint());
        } else if (openIDMetadata.getUserInfoEndpoint() != null) {
            additionalHeaders.put("Authorization", bearerAuth(openIDMetadata.getAccessToken()));
            // get
            responseJson = HttpClient.get(openIDMetadata.getUserInfoEndpoint(), additionalHeaders);
            if (responseJson != null) {
                result = GSON.fromJson(responseJson, IntrospectionResponse.class);
                if (result.getSub() != null) {
                    result.setClientId(openIDMetadata.getClientId());
                    result.setActive(true);
                    if (result.getUsername() == null)
                        result.setUsername(result.getSub());
                    if (result.getExp() == null) {
                        if (openIDMetadata.getTokenExpirationTime() > 0)
                            result.setExp(convertMillisecondsToSeconds(openIDMetadata.getTokenExpirationTime()));
                        else {
                            try {
                                Claims claims = getJwtClaims(openIDMetadata.getAccessToken(), openIDMetadata.getJwksEndpoint());
                                if (claims.getExpiration() != null) {
                                    result.setExp(convertMillisecondsToSeconds(claims.getExpiration().getTime()));
                                }
                            } catch (Exception e) {
                                LogManager.logDetail(LogConstants.CTX_RUNTIME, RuntimePlugin.Util.gs(RuntimePlugin.Event.TEIID40203, "It's not possible to read \"exp\" claim in JWT of access token"));
                            }
                        }
                    }
                }
            } else if (throwException)
                throw new SessionServiceException(RuntimePlugin.Event.TEIID40201, "Empty response from " + openIDMetadata.getUserInfoEndpoint());
        } else if (throwException)
            throw new SessionServiceException(RuntimePlugin.Event.TEIID40201, "Error checking token, introspection and userInfo endpoints are empty");

        return result;
    }

    private long convertMillisecondsToSeconds(long milliseconds) {
        return TimeUnit.MILLISECONDS.toSeconds(milliseconds - System.currentTimeMillis());
    }

    private Subject createSubject(OpenIDMetadata openIDMetadata, String clientId, String username) throws SessionServiceException {
        String token = openIDMetadata.getAccessToken();
        if (token == null || token.isEmpty()) {
            return null;
        }

        // extract roles from the token
        List<String> roles = null;
        try {
            String jwksEndpoint = openIDMetadata.getJwksEndpoint();
            Claims claims = getJwtClaims(token, jwksEndpoint);

            // getting username
            if (claims.containsKey("preferred_username")) {
                username = claims.get("preferred_username", String.class);
            } else if (claims.getSubject() != null)
                username = claims.getSubject();

            /*if (username != null && username.contains("@"))
                username = username.substring(0, username.indexOf("@"));*/

            // load users and roles from the provider
            String provider = openIDMetadata.getProvider();
            Properties props = new Properties();
            boolean loadUsers = true;
            String baseMessage = "The %s property is empty, users/roles won't be loaded from the %s provider.";
            if (openIDMetadata.getServerUrl() == null || openIDMetadata.getServerUrl().isEmpty()) {
                LogManager.logInfo(LogConstants.CTX_RUNTIME, String.format(baseMessage, "serverUrl", provider));
                loadUsers = false;
            } else
                props.setProperty(DV_AUTH_SERVER_URL, openIDMetadata.getServerUrl());

            if (openIDMetadata.getRealm() == null || openIDMetadata.getRealm().isEmpty()) {
                LogManager.logInfo(LogConstants.CTX_RUNTIME, String.format(baseMessage, "realm", provider));
                loadUsers = false;
            } else
                props.setProperty(DV_AUTH_REALM, openIDMetadata.getRealm());

            if (loadUsers) {
                props.setProperty(DV_AUTH_CLIENT_ID, openIDMetadata.getClientId());
                props.setProperty(DV_AUTH_CLIENT_SECRET, openIDMetadata.getClientSecret());
                DVCaller.call(241, provider, props, username);
            }

            // getting roles
            if (openIDMetadata.isUseDvRoles()) {
                // let's try to get DV roles
                roles = (List<String>) DVCaller.callWithThrow(SessionServiceException.class, 237, getDomainName(username), null);
                // if there is still no the domain user let's try to get a role by no domain name
                if (roles == null)
                    roles = (List<String>) DVCaller.callWithThrow(SessionServiceException.class, 237, username, null);
            } else {
                if (claims.containsKey("resource_access")) {
                    Map<String, Object> resourceAccess = claims.get("resource_access", Map.class);
                    Map<String, Object> account = null;
                    if (resourceAccess.containsKey(clientId))
                        account = (Map<String, Object>) resourceAccess.get(clientId);
                    else if (resourceAccess.containsKey("account"))
                        account = (Map<String, Object>) resourceAccess.get("account");
                    roles = getRoles(account);
                } else if (claims.containsKey("roles")) {
                    roles = getRoles(claims);
                }

                roles = (List<String>) DVCaller.callWithThrow(SessionServiceException.class, 237, getDomainName(username), roles);
            }
        } catch (SessionServiceException e) {
            throw e;
        } catch (Exception e) {
            LogManager.logError(LogConstants.CTX_RUNTIME, e, "Error parsing token");
        }

        Subject subject = new Subject();
        Principal userPrincipal = new SimplePrincipal(getDomainName(username));
        subject.getPrincipals().add(userPrincipal);

        SimpleGroup rolesGroup = new SimpleGroup("Roles");
        if (roles != null) {
            roles.forEach(role -> rolesGroup.addMember(new SimplePrincipal(role)));
        }
        subject.getPrincipals().add(rolesGroup);

        SimpleGroup callerGroup = new SimpleGroup("CallerPrincipal");
        callerGroup.addMember(userPrincipal);
        subject.getPrincipals().add(callerGroup);

        return subject;
    }

    private String getDomainName(String name) {
        return name + DOMAIN;
    }

    private Claims getJwtClaims(String token, String jwksEndpoint) throws SessionServiceException {
        try {
            JwtParser jwtParser = Jwts.parserBuilder().build();
            if (jwtParser.isSigned(token)) {
                JwksResponse jwksResponse = null;
                if (jwksEndpoint != null) {
                    String responseJson = HttpClient.get(jwksEndpoint);
                    if (responseJson != null)
                        jwksResponse = GSON.fromJson(responseJson, JwksResponse.class);
                }
                if (jwksResponse != null && jwksResponse.getJwksKeys() != null) {
                    Jws<Claims> parsedToken = null;
                    for (JwksKey jwksKey : jwksResponse.getJwksKeys()) {
                        try {
                            // at the moment we support only RSA algorithm
                            if (jwksKey.getKeyType().equalsIgnoreCase("RSA")) {
                                KeyFactory keyFactory = KeyFactory.getInstance(jwksKey.getKeyType());
                                BigInteger modulus = new BigInteger(1, Base64.getUrlDecoder().decode(jwksKey.getModulus()));
                                BigInteger exponent = new BigInteger(1, Base64.getUrlDecoder().decode(jwksKey.getExponent()));
                                RSAPublicKey pubKey = (RSAPublicKey) keyFactory.generatePublic(new RSAPublicKeySpec(modulus, exponent));

                                jwtParser = Jwts.parserBuilder().setSigningKey(pubKey).build();
                                parsedToken = jwtParser.parseClaimsJws(token);
                                break;
                            }
                        } catch (Exception e) {
                            LogManager.logDetail(LogConstants.CTX_RUNTIME, e, "Error verifying JWS token");
                        }
                    }
                    if (parsedToken == null)
                        throw new SessionServiceException(RuntimePlugin.Event.TEIID40201, "Error verifying JWS token, no valid JWS keys found");
                    return parsedToken.getBody();
                } else {
                    // skip verification
                    String[] splitToken = token.split("\\" + JwtParser.SEPARATOR_CHAR);
                    token = splitToken[0] + JwtParser.SEPARATOR_CHAR + splitToken[1] + JwtParser.SEPARATOR_CHAR;
                }
            }
            Jwt<Header, Claims> parsedToken = jwtParser.parseClaimsJwt(token);
            return parsedToken.getBody();
        } catch (Exception e) {
            throw new SessionServiceException(e, "Access token must conform to the JWT standard and contain information about an entity in the form of claims");
        }
    }

    private List<String> getRoles(Map<String, Object> account) {
        List<String> roles = null;
        if (account != null && account.containsKey("roles")) {
            roles = (List<String>) account.get("roles");
            roles = roles.stream().map(s -> getDomainName(s)).collect(Collectors.toList());
        }
        return roles;
    }

    private static TokenResponse getToken(String tokenEndpoint, String grantType, String clientId, String clientSecret,
                                          Map<String, String> additionalParams) {

        Map<String, String> additionalHeaders = new HashMap<>();
        StringBuilder params = getAuthorizationParams(new StringBuilder(), clientId, clientSecret);

        // client auth if needed
        /*if (clientId != null && clientSecret != null) {
            additionalHeaders.put("Authorization", basicAuth(clientId, clientSecret));
        }*/

        // build params
        if (params.length() > 0)
            params.append("&");
        params.append("grant_type=").append(grantType);
        if (additionalParams != null) {
            for (Map.Entry<String, String> entry : additionalParams.entrySet()) {
                params.append("&")
                        .append(entry.getKey())
                        .append("=")
                        .append(entry.getValue());
            }
        }

        // post
        String responseJson = HttpClient.post(tokenEndpoint, params.toString(), additionalHeaders);

        if (responseJson == null) {
            throw new RuntimeException("Empty response from " + tokenEndpoint);
        }
        return GSON.fromJson(responseJson, TokenResponse.class);
    }

    private static StringBuilder getAuthorizationParams(StringBuilder params, String clientId, String clientSecret) {
        if (clientId != null && clientSecret != null) {
            if (params.length() > 0)
                params.append("&");
            params.append("client_id=").append(clientId);
            params.append("&");
            params.append("client_secret=").append(clientSecret);
        }

        return params;
    }

    private static String basicAuth(String clientId, String clientSecret) {
        byte[] credentials = (clientId + ":" + clientSecret).getBytes(StandardCharsets.UTF_8);
        String encodedCredentials = Base64.getEncoder().encodeToString(credentials);
        return "Basic " + encodedCredentials;
    }

    private static String bearerAuth(String token) {
        return "Bearer " + token;
    }
}